import unittest
import neet.boolean as bnet
import numpy as np


class TestWTNetwork(unittest.TestCase):
    def test_is_network(self):
        from neet.interfaces import is_network
        self.assertTrue(is_network(bnet.WTNetwork))
        self.assertTrue(is_network(bnet.WTNetwork([[1]])))

    def test_is_fixed_sized(self):
        from neet.interfaces import is_fixed_sized
        self.assertTrue(is_fixed_sized(bnet.WTNetwork))
        self.assertTrue(is_fixed_sized(bnet.WTNetwork([[1]])))

    def test_init_failed(self):
        with self.assertRaises(ValueError):
            bnet.WTNetwork(None)

        with self.assertRaises(ValueError):
            bnet.WTNetwork('a')

        with self.assertRaises(ValueError):
            bnet.WTNetwork(0)

        with self.assertRaises(ValueError):
            bnet.WTNetwork(-1)

        with self.assertRaises(ValueError):
            bnet.WTNetwork([[1]], 'a')

        with self.assertRaises(ValueError):
            bnet.WTNetwork([[1]], 0)

        with self.assertRaises(ValueError):
            bnet.WTNetwork([[1]], -1)

        with self.assertRaises(ValueError):
            bnet.WTNetwork([])

        with self.assertRaises(ValueError):
            bnet.WTNetwork([[]])

        with self.assertRaises(ValueError):
            bnet.WTNetwork([[1]], [])

        with self.assertRaises(ValueError):
            bnet.WTNetwork([[1]], [1, 2])

    def test_init_weights(self):
        net = bnet.WTNetwork([[1]])
        self.assertEqual(1, net.size)
        self.assertTrue(np.array_equal([[1]], net.weights))
        self.assertTrue(np.array_equal([0], net.thresholds))
        self.assertIsNone(net.names)

        net = bnet.WTNetwork([[1, 0], [0, 1]])
        self.assertEqual(2, net.size)
        self.assertTrue(np.array_equal([[1, 0], [0, 1]], net.weights))
        self.assertTrue(np.array_equal([0, 0], net.thresholds))
        self.assertIsNone(net.names)

        net = bnet.WTNetwork([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        self.assertEqual(3, net.size)
        self.assertTrue(np.array_equal(
            [[1, 0, 0], [0, 1, 0], [0, 0, 1]], net.weights))
        self.assertTrue(np.array_equal([0, 0, 0], net.thresholds))
        self.assertIsNone(net.names)

        net = bnet.WTNetwork(1)
        self.assertEqual(1, net.size)
        self.assertTrue(np.array_equal([[0]], net.weights))
        self.assertTrue(np.array_equal([0], net.thresholds))
        self.assertIsNone(net.names)

        net = bnet.WTNetwork(2)
        self.assertEqual(2, net.size)
        self.assertTrue(np.array_equal([[0, 0], [0, 0]], net.weights))
        self.assertTrue(np.array_equal([0, 0], net.thresholds))
        self.assertIsNone(net.names)

    def test_init_weights_thresholds(self):
        net = bnet.WTNetwork([[1]], [1])
        self.assertEqual(1, net.size)
        self.assertTrue(np.array_equal([[1]], net.weights))
        self.assertTrue(np.array_equal([1], net.thresholds))
        self.assertIsNone(net.names)

        net = bnet.WTNetwork([[1, 0], [0, 1]], [1, 2])
        self.assertEqual(2, net.size)
        self.assertTrue(np.array_equal([[1, 0], [0, 1]], net.weights))
        self.assertTrue(np.array_equal([1, 2], net.thresholds))
        self.assertIsNone(net.names)

        net = bnet.WTNetwork([[1, 0, 0], [0, 1, 0], [0, 0, 1]], [1, 2, 3])
        self.assertEqual(3, net.size)
        self.assertTrue(np.array_equal(
            [[1, 0, 0], [0, 1, 0], [0, 0, 1]], net.weights))
        self.assertTrue(np.array_equal([1, 2, 3], net.thresholds))
        self.assertIsNone(net.names)

    def test_init_names(self):
        with self.assertRaises(TypeError):
            bnet.WTNetwork([[1]], names=5)

        with self.assertRaises(ValueError):
            bnet.WTNetwork([[1]], names=["A", "B"])

        with self.assertRaises(ValueError):
            bnet.WTNetwork([[1]], names=["A", "B"])

        net = bnet.WTNetwork([[1]], names=["A"])
        self.assertEqual(['A'], net.names)

        net = bnet.WTNetwork([[1, 0], [1, 1]], names=['A', 'B'])
        self.assertEqual(['A', 'B'], net.names)

        net = bnet.WTNetwork([[1]], names="A")
        self.assertEqual(['A'], net.names)

        net = bnet.WTNetwork([[1, 0], [1, 1]], names="AB")
        self.assertEqual(['A', 'B'], net.names)

    def test_init_thresholds(self):
        with self.assertRaises(TypeError):
            bnet.WTNetwork([[1]], theta=5)

        net = bnet.WTNetwork([[1]])
        self.assertEqual(bnet.WTNetwork.split_threshold, net.theta)

        net = bnet.WTNetwork([[1]], theta=bnet.WTNetwork.negative_threshold)
        self.assertEqual(bnet.WTNetwork.negative_threshold, net.theta)

        net = bnet.WTNetwork([[1]], theta=bnet.WTNetwork.positive_threshold)
        self.assertEqual(bnet.WTNetwork.positive_threshold, net.theta)

    def test_state_space(self):
        net = bnet.WTNetwork([[1]])
        self.assertEqual(2, len(list(net.state_space())))
        net = bnet.WTNetwork([[1, 0], [1, 1]])
        self.assertEqual(4, len(list(net.state_space())))
        net = bnet.WTNetwork([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        self.assertEqual(8, len(list(net.state_space())))

    def test_update_empty_states(self):
        net = bnet.WTNetwork([[1, 0], [1, 1]])
        with self.assertRaises(ValueError):
            net.update([])

    def test_update_invalid_states(self):
        net = bnet.WTNetwork([[1, 0], [1, 1]])
        with self.assertRaises(ValueError):
            net.update([-1, 0])

        with self.assertRaises(ValueError):
            net.update([1, -1])

        with self.assertRaises(ValueError):
            net.update([2, 0])

        with self.assertRaises(ValueError):
            net.update([1, 2])

        with self.assertRaises(ValueError):
            net.update([[1], [0]])

        with self.assertRaises(ValueError):
            net.update("101")

    def test_update_invalid_index(self):
        net = bnet.WTNetwork([[1, 0], [1, 1]])
        with self.assertRaises(IndexError):
            net.update([0, 0], 2)

    def test_update(self):
        net = bnet.WTNetwork([[1, 0], [-1, 1]], [0.5, 0.0])

        self.assertEqual(bnet.WTNetwork.split_threshold, net.theta)

        xs = [0, 0]
        self.assertEqual([0, 0], net.update(xs))
        self.assertEqual([0, 0], xs)

        xs = [1, 0]
        self.assertEqual([1, 0], net.update(xs))
        self.assertEqual([1, 0], xs)

        xs = [0, 1]
        self.assertEqual([0, 1], net.update(xs))
        self.assertEqual([0, 1], xs)

        xs = [1, 1]
        self.assertEqual([1, 1], net.update(xs))
        self.assertEqual([1, 1], xs)

        net.theta = bnet.WTNetwork.negative_threshold

        xs = [0, 0]
        self.assertEqual([0, 0], net.update(xs))
        self.assertEqual([0, 0], xs)

        xs = [1, 0]
        self.assertEqual([1, 0], net.update(xs))
        self.assertEqual([1, 0], xs)

        xs = [0, 1]
        self.assertEqual([0, 1], net.update(xs))
        self.assertEqual([0, 1], xs)

        xs = [1, 1]
        self.assertEqual([1, 0], net.update(xs))
        self.assertEqual([1, 0], xs)

        net.theta = bnet.WTNetwork.positive_threshold

        xs = [0, 0]
        self.assertEqual([0, 1], net.update(xs))
        self.assertEqual([0, 1], xs)

        xs = [1, 0]
        self.assertEqual([1, 0], net.update(xs))
        self.assertEqual([1, 0], xs)

        xs = [0, 1]
        self.assertEqual([0, 1], net.update(xs))
        self.assertEqual([0, 1], xs)

        xs = [1, 1]
        self.assertEqual([1, 1], net.update(xs))
        self.assertEqual([1, 1], xs)

    def test_update_index(self):
        net = bnet.WTNetwork([[1, 0], [-1, 1]], [0.5, 0.0])

        self.assertEqual(bnet.WTNetwork.split_threshold, net.theta)

        xs = [0, 0]
        self.assertEqual([0, 0], net.update(xs, 0))
        self.assertEqual([0, 0], net.update(xs, 1))
        self.assertEqual([0, 0], xs)

        xs = [1, 0]
        self.assertEqual([1, 0], net.update(xs, 0))
        self.assertEqual([1, 0], net.update(xs, 1))
        self.assertEqual([1, 0], xs)

        xs = [0, 1]
        self.assertEqual([0, 1], net.update(xs, 0))
        self.assertEqual([0, 1], net.update(xs, 1))
        self.assertEqual([0, 1], xs)

        xs = [1, 1]
        self.assertEqual([1, 1], net.update(xs, 0))
        self.assertEqual([1, 1], net.update(xs, 1))
        self.assertEqual([1, 1], xs)

        net.theta = bnet.WTNetwork.negative_threshold

        xs = [0, 0]
        self.assertEqual([0, 0], net.update(xs, 0))
        self.assertEqual([0, 0], net.update(xs, 1))
        self.assertEqual([0, 0], xs)

        xs = [1, 0]
        self.assertEqual([1, 0], net.update(xs, 0))
        self.assertEqual([1, 0], net.update(xs, 1))
        self.assertEqual([1, 0], xs)

        xs = [0, 1]
        self.assertEqual([0, 1], net.update(xs, 0))
        self.assertEqual([0, 1], net.update(xs, 1))
        self.assertEqual([0, 1], xs)

        xs = [1, 1]
        self.assertEqual([1, 1], net.update(xs, 0))
        self.assertEqual([1, 0], net.update(xs, 1))
        self.assertEqual([1, 0], xs)

        net.theta = bnet.WTNetwork.positive_threshold

        xs = [0, 0]
        self.assertEqual([0, 0], net.update(xs, 0))
        self.assertEqual([0, 1], net.update(xs, 1))
        self.assertEqual([0, 1], xs)

        xs = [1, 0]
        self.assertEqual([1, 0], net.update(xs, 0))
        self.assertEqual([1, 0], net.update(xs, 1))
        self.assertEqual([1, 0], xs)

        xs = [0, 1]
        self.assertEqual([0, 1], net.update(xs, 0))
        self.assertEqual([0, 1], net.update(xs, 1))
        self.assertEqual([0, 1], xs)

        xs = [1, 1]
        self.assertEqual([1, 1], net.update(xs, 0))
        self.assertEqual([1, 1], net.update(xs, 1))
        self.assertEqual([1, 1], xs)

    def test_user_defined_threshold(self):
        def reverse_negative(values, states):
            if isinstance(values, list) or isinstance(values, np.ndarray):
                for i, x in enumerate(values):
                    if x <= 0:
                        states[i] = 1
                    else:
                        states[i] = 0
                return states
            else:
                if values <= 0:
                    return 1
                else:
                    return 0

        net = bnet.WTNetwork([[1, 0], [-1, 1]], [0.5, 0.0],
                             theta=reverse_negative)
        xs = [0, 0]
        self.assertEqual([1, 1], net.update(xs))
        self.assertEqual([1, 1], xs)

        xs = [0, 0]
        self.assertEqual([1, 0], net.update(xs, 0))
        self.assertEqual([1, 0], xs)

        xs = [0, 0]
        self.assertEqual([0, 1], net.update(xs, 1))
        self.assertEqual([0, 1], xs)

        xs = [1, 0]
        self.assertEqual([0, 1], net.update(xs))
        self.assertEqual([0, 1], xs)

        xs = [1, 0]
        self.assertEqual([0, 0], net.update(xs, 0))
        self.assertEqual([0, 0], xs)

        xs = [1, 0]
        self.assertEqual([1, 1], net.update(xs, 1))
        self.assertEqual([1, 1], xs)

        xs = [0, 1]
        self.assertEqual([1, 0], net.update(xs))
        self.assertEqual([1, 0], xs)

        xs = [0, 1]
        self.assertEqual([1, 1], net.update(xs, 0))
        self.assertEqual([1, 1], xs)

        xs = [0, 1]
        self.assertEqual([0, 0], net.update(xs, 1))
        self.assertEqual([0, 0], xs)

        xs = [1, 1]
        self.assertEqual([0, 1], net.update(xs))
        self.assertEqual([0, 1], xs)

        xs = [1, 1]
        self.assertEqual([0, 1], net.update(xs, 0))
        self.assertEqual([0, 1], xs)

        xs = [1, 1]
        self.assertEqual([1, 1], net.update(xs, 1))
        self.assertEqual([1, 1], xs)

    def test_fission_yeast(self):
        net = bnet.WTNetwork(
            [[-1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
             [0.0, 0.0, -1.0, -1.0, -1.0, 0.0, 0.0, 0.0, 0.0],
             [-1.0, -1.0, 0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 1.0],
             [-1.0, -1.0, 0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 1.0],
             [0.0, 0.0, 0.0, 0.0, -1.0, 1.0, 0.0, 0.0, 0.0],
             [0.0, 0.0, -1.0, -1.0, -1.0, 0.0, -1.0, 1.0, 0.0],
             [0.0, -1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0],
             [0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -1.0],
             [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, -1.0]],
            [0.0, -0.5, 0.0, 0.0, 0.0, 0.5, 0.0, 0.0, 0.0])

        self.assertEqual(9, net.size)
        self.assertEqual(512, len(list(net.state_space())))

        init = [1, 0, 1, 1, 0, 0, 1, 0, 0]
        bio_sequence = [[0, 0, 0, 0, 0, 0, 1, 0, 0],
                        [0, 1, 0, 0, 0, 0, 1, 0, 0],
                        [0, 1, 0, 0, 0, 0, 0, 1, 0],
                        [0, 1, 0, 0, 0, 1, 0, 1, 0],
                        [0, 1, 0, 0, 1, 1, 0, 1, 0],
                        [0, 0, 0, 0, 1, 0, 0, 1, 1],
                        [0, 0, 1, 1, 0, 0, 1, 0, 1],
                        [0, 0, 1, 1, 0, 0, 1, 0, 0]]

        for expected in bio_sequence:
            self.assertEqual(expected, net.update(init))

    def test_fission_yeast_numpy(self):
        net = bnet.WTNetwork(
            [[-1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
             [0.0, 0.0, -1.0, -1.0, -1.0, 0.0, 0.0, 0.0, 0.0],
             [-1.0, -1.0, 0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 1.0],
             [-1.0, -1.0, 0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 1.0],
             [0.0, 0.0, 0.0, 0.0, -1.0, 1.0, 0.0, 0.0, 0.0],
             [0.0, 0.0, -1.0, -1.0, -1.0, 0.0, -1.0, 1.0, 0.0],
             [0.0, -1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0],
             [0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -1.0],
             [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, -1.0]],
            [0.0, -0.5, 0.0, 0.0, 0.0, 0.5, 0.0, 0.0, 0.0])

        self.assertEqual(9, net.size)
        self.assertEqual(512, len(list(net.state_space())))

        init = np.asarray([1, 0, 1, 1, 0, 0, 1, 0, 0])
        bio_sequence = np.asarray([[0, 0, 0, 0, 0, 0, 1, 0, 0],
                                   [0, 1, 0, 0, 0, 0, 1, 0, 0],
                                   [0, 1, 0, 0, 0, 0, 0, 1, 0],
                                   [0, 1, 0, 0, 0, 1, 0, 1, 0],
                                   [0, 1, 0, 0, 1, 1, 0, 1, 0],
                                   [0, 0, 0, 0, 1, 0, 0, 1, 1],
                                   [0, 0, 1, 1, 0, 0, 1, 0, 1],
                                   [0, 0, 1, 1, 0, 0, 1, 0, 0]])

        for expected in bio_sequence:
            self.assertTrue(np.array_equal(expected, net.update(init)))

    def test_split_threshold(self):
        xs = [0, 0, 0]
        self.assertEqual(
            [1, 0, 0], bnet.WTNetwork.split_threshold([1, -1, 0], xs))
        self.assertEqual([1, 0, 0], xs)

        xs = [1, 1, 1]
        self.assertEqual(
            [1, 0, 1], bnet.WTNetwork.split_threshold([1, -1, 0], xs))
        self.assertEqual([1, 0, 1], xs)

    def test_split_threshold_scalar(self):
        test = {
            (1, 0): 1,
            (0, 0): 0,
            (-1, 0): 0,
            (1, 1): 1,
            (0, 1): 1,
            (-1, 1): 0,
        }
        for x, s in test:
            self.assertEqual(
                test[(x, s)], bnet.WTNetwork.split_threshold(x, s))

    def test_negative_threshold(self):
        xs = [0, 0, 0]
        self.assertEqual(
            [1, 0, 0], bnet.WTNetwork.negative_threshold([1, -1, 0], xs))
        self.assertEqual([1, 0, 0], xs)

        xs = [1, 1, 1]
        self.assertEqual(
            [1, 0, 0], bnet.WTNetwork.negative_threshold([1, -1, 0], xs))
        self.assertEqual([1, 0, 0], xs)

    def test_negative_threshold_scalar(self):
        test = {
            (1, 0): 1,
            (0, 0): 0,
            (-1, 0): 0,
            (1, 1): 1,
            (0, 1): 0,
            (-1, 1): 0,
        }
        for x, s in test:
            self.assertEqual(
                test[(x, s)], bnet.WTNetwork.negative_threshold(x, s))

    def test_positive_threshold(self):
        xs = [0, 0, 0]
        self.assertEqual(
            [1, 0, 1], bnet.WTNetwork.positive_threshold([1, -1, 0], xs))
        self.assertEqual([1, 0, 1], xs)

        xs = [1, 1, 1]
        self.assertEqual(
            [1, 0, 1], bnet.WTNetwork.positive_threshold([1, -1, 0], xs))
        self.assertEqual([1, 0, 1], xs)

    def test_positive_threshold_scalar(self):
        test = {
            (1, 0): 1,
            (0, 0): 1,
            (-1, 0): 0,
            (1, 1): 1,
            (0, 1): 1,
            (-1, 1): 0,
        }
        for x, s in test:
            self.assertEqual(
                test[(x, s)], bnet.WTNetwork.positive_threshold(x, s))

    def test_update_pin_none(self):
        net = bnet.WTNetwork([[1, 0], [-1, 1]], [0.5, 0.0],
                             theta=bnet.WTNetwork.positive_threshold)
        xs = [0, 0]
        self.assertEqual([0, 1], net.update(xs, pin=None))
        xs = [0, 0]
        self.assertEqual([0, 1], net.update(xs, pin=[]))

    def test_update_pin_index_clash(self):
        net = bnet.WTNetwork([[1, 0], [-1, 1]], [0.5, 0.0],
                             theta=bnet.WTNetwork.positive_threshold)
        with self.assertRaises(ValueError):
            net.update([0, 0], index=0, pin=[1])
        with self.assertRaises(ValueError):
            net.update([0, 0], index=1, pin=[1])
        with self.assertRaises(ValueError):
            net.update([0, 0], index=1, pin=[0, 1])

    def test_update_pin(self):
        net = bnet.WTNetwork([[1, 0], [-1, 1]], [0.5, 0.0])

        net.theta = bnet.WTNetwork.negative_threshold
        xs = [1, 1]
        self.assertEqual([1, 0], net.update(xs, pin=[0]))

        net.theta = bnet.WTNetwork.positive_threshold
        xs = [0, 0]
        self.assertEqual([0, 0], net.update(xs, pin=[1]))

    def test_pinning_s_pombe(self):
        from neet.boolean.examples import s_pombe
        self.assertEqual(
            [0, 0, 0, 0, 0, 0, 0, 0, 0],
            s_pombe.update([0, 0, 0, 0, 1, 0, 0, 0, 0], pin=[-1])
        )
        self.assertEqual(
            [0, 0, 1, 1, 0, 0, 1, 0, 0],
            s_pombe.update([0, 0, 0, 0, 0, 0, 0, 0, 1], pin=[1])
        )
        self.assertEqual(
            [0, 0, 0, 0, 0, 0, 1, 0, 0],
            s_pombe.update([0, 0, 0, 0, 0, 0, 0, 0, 1], pin=range(1, 4))
        )
        self.assertEqual(
            [0, 0, 0, 0, 0, 0, 1, 0, 1],
            s_pombe.update([0, 0, 0, 0, 0, 0, 0, 0, 1], pin=[1, 2, 3, -1])
        )

    def test_update_values_none(self):
        net = bnet.WTNetwork([[1, 0], [-1, 1]], [0.5, 0.0],
                             theta=bnet.WTNetwork.positive_threshold)
        xs = [0, 0]
        self.assertEqual([0, 1], net.update(xs, values=None))
        xs = [0, 0]
        self.assertEqual([0, 1], net.update(xs, values={}))

    def test_update_invalid_values(self):
        net = bnet.WTNetwork([[1, 0], [-1, 1]], [0.5, 0.0],
                             theta=bnet.WTNetwork.positive_threshold)
        with self.assertRaises(ValueError):
            net.update([0, 0], values={0: 2})
        with self.assertRaises(ValueError):
            net.update([0, 0], values={0: -1})

    def test_update_pin_invalid_indicies(self):
        net = bnet.WTNetwork([[1, 0], [-1, 1]], [0.5, 0.0],
                             theta=bnet.WTNetwork.positive_threshold)
        with self.assertRaises(IndexError):
            net.update([0, 0], values={-3: 0})
        with self.assertRaises(IndexError):
            net.update([0, 0], values={2: 0})

    def test_update_values_index_clash(self):
        net = bnet.WTNetwork([[1, 0], [-1, 1]], [0.5, 0.0],
                             theta=bnet.WTNetwork.positive_threshold)
        with self.assertRaises(ValueError):
            net.update([0, 0], index=0, values={0: 1})
        with self.assertRaises(ValueError):
            net.update([0, 0], index=1, values={1: 0})
        with self.assertRaises(ValueError):
            net.update([0, 0], index=1, values={0: 0, 1: 0})

    def test_update_values_pin_clash(self):
        net = bnet.WTNetwork([[1, 0], [-1, 1]], [0.5, 0.0],
                             theta=bnet.WTNetwork.positive_threshold)
        with self.assertRaises(ValueError):
            net.update([0, 0], pin=[0], values={0: 1})
        with self.assertRaises(ValueError):
            net.update([0, 0], pin=[1], values={1: 0})
        with self.assertRaises(ValueError):
            net.update([0, 0], pin=[1], values={0: 0, 1: 0})
        with self.assertRaises(ValueError):
            net.update([0, 0], pin=[1, 0], values={0: 0})

    def test_update_values(self):
        net = bnet.WTNetwork([[1, 0], [-1, 1]], [0.5, 0.0],
                             theta=bnet.WTNetwork.negative_threshold)

        xs = [1, 1]
        self.assertEqual([1, 1], net.update(xs, values={1: 1}))

        net.theta = bnet.WTNetwork.positive_threshold
        xs = [0, 0]
        self.assertEqual([1, 1], net.update(xs, values={0: 1}))

    def test_values_s_pombe(self):
        from neet.boolean.examples import s_pombe
        self.assertEqual(
            [0, 1, 0, 0, 1, 0, 0, 0, 0],
            s_pombe.update([0, 0, 0, 0, 1, 0, 0, 0, 0],
                           values={1: 1, 4: 1, -1: 0})
        )
        self.assertEqual(
            [0, 1, 0, 1, 0, 0, 1, 1, 0],
            s_pombe.update([0, 0, 0, 0, 0, 0, 0, 0, 1], values={2: 0, -2: 1})
        )
        self.assertEqual(
            [0, 0, 0, 0, 0, 0, 1, 0, 0],
            s_pombe.update([0, 0, 0, 0, 0, 0, 0, 0, 1],
                           values={1: 0, 2: 0, 3: 0})
        )

    def test_has_metadata(self):
        net = bnet.WTNetwork([[1]])
        self.assertTrue(hasattr(net, 'metadata'))
        self.assertEqual(type(net.metadata), dict)

    def test_neighbors_in_split_threshold(self):

        net = bnet.WTNetwork(
            [[-1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
             [0.0, 0.0, -1.0, -1.0, -1.0, 0.0, 0.0, 0.0, 0.0],
             [-1.0, -1.0, 0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 1.0],
             [-1.0, -1.0, 0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 1.0],
             [0.0, 0.0, 0.0, 0.0, -1.0, 1.0, 0.0, 0.0, 0.0],
             [0.0, 0.0, -1.0, -1.0, -1.0, 0.0, -1.0, 1.0, 0.0],
             [0.0, -1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0],
             [0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -1.0],
             [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, -1.0]],
            [0.0, -0.5, 0.0, 0.0, 0.0, 0.5, 0.0, 0.0, 0.0])

        self.assertEqual(net.neighbors_in(2), set([0, 1, 2, 5, 8]))

        with self.assertRaises(IndexError):
            self.assertEqual(net.neighbors_in(2.0))

        with self.assertRaises(IndexError):
            self.assertEqual(net.neighbors_in('2'))

    def test_neighbors_out_split_threshold(self):

        net = bnet.WTNetwork(
            [[-1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
             [0.0, 0.0, -1.0, -1.0, -1.0, 0.0, 0.0, 0.0, 0.0],
             [-1.0, -1.0, 0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 1.0],
             [-1.0, -1.0, 0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 1.0],
             [0.0, 0.0, 0.0, 0.0, -1.0, 1.0, 0.0, 0.0, 0.0],
             [0.0, 0.0, -1.0, -1.0, -1.0, 0.0, -1.0, 1.0, 0.0],
             [0.0, -1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0],
             [0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -1.0],
             [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, -1.0]],
            [0.0, -0.5, 0.0, 0.0, 0.0, 0.5, 0.0, 0.0, 0.0])

        self.assertEqual(net.neighbors_out(2), set([1, 2, 5]))

    def test_neighbors_in_positive_threshold(self):

        net = bnet.WTNetwork(
            [[-1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
             [0.0, 0.0, -1.0, -1.0, -1.0, 0.0, 0.0, 0.0, 0.0],
             [-1.0, -1.0, 0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 1.0],
             [-1.0, -1.0, 0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 1.0],
             [0.0, 0.0, 0.0, 0.0, -1.0, 1.0, 0.0, 0.0, 0.0],
             [0.0, 0.0, -1.0, -1.0, -1.0, 0.0, -1.0, 1.0, 0.0],
             [0.0, -1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0],
             [0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -1.0],
             [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, -1.0]],
            [0.0, -0.5, 0.0, 0.0, 0.0, 0.5, 0.0, 0.0, 0.0],
            theta=bnet.WTNetwork.positive_threshold)

        self.assertEqual(net.neighbors_in(2), set([0, 1, 5, 8]))

        with self.assertRaises(IndexError):
            self.assertEqual(net.neighbors_in(2.0))

        with self.assertRaises(IndexError):
            self.assertEqual(net.neighbors_in('2'))

    def test_neighbors_out_negative_threshold(self):

        net = bnet.WTNetwork(
            [[-1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
             [0.0, 0.0, -1.0, -1.0, -1.0, 0.0, 0.0, 0.0, 0.0],
             [-1.0, -1.0, 0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 1.0],
             [-1.0, -1.0, 0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 1.0],
             [0.0, 0.0, 0.0, 0.0, -1.0, 1.0, 0.0, 0.0, 0.0],
             [0.0, 0.0, -1.0, -1.0, -1.0, 0.0, -1.0, 1.0, 0.0],
             [0.0, -1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0],
             [0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -1.0],
             [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, -1.0]],
            [0.0, -0.5, 0.0, 0.0, 0.0, 0.5, 0.0, 0.0, 0.0],
            theta=bnet.WTNetwork.negative_threshold)

        self.assertEqual(net.neighbors_out(2), set([1, 5]))

    def test_neighbors_both_split_threshold(self):

        net = bnet.WTNetwork(
            [[-1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
             [0.0, 0.0, -1.0, -1.0, -1.0, 0.0, 0.0, 0.0, 0.0],
             [-1.0, -1.0, 0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 1.0],
             [-1.0, -1.0, 0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 1.0],
             [0.0, 0.0, 0.0, 0.0, -1.0, 1.0, 0.0, 0.0, 0.0],
             [0.0, 0.0, -1.0, -1.0, -1.0, 0.0, -1.0, 1.0, 0.0],
             [0.0, -1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0],
             [0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -1.0],
             [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, -1.0]],
            [0.0, -0.5, 0.0, 0.0, 0.0, 0.5, 0.0, 0.0, 0.0])

        self.assertEqual(net.neighbors(2), set([0, 1, 2, 5, 8]))

    def test_to_networkx_graph_names(self):
        from neet.boolean.examples import s_pombe

        nx_net = s_pombe.to_networkx_graph(labels='names')
        self.assertEqual(set(nx_net), set(s_pombe.names))

    def test_to_networkx_graph_names_fail(self):
        net = bnet.WTNetwork([[1, 0], [0, 1]])

        with self.assertRaises(ValueError):
            net.to_networkx_graph(labels='names')

    def test_to_networkx_metadata(self):
        from neet.boolean.examples import s_pombe

        nx_net = s_pombe.to_networkx_graph(labels='indices')

        self.assertEqual(nx_net.graph['name'], 's_pombe')
        self.assertEqual(nx_net.graph['name'], s_pombe.metadata['name'])

    # def test_draw(self):
    #     net = bnet.LogicNetwork([((0,), {'0'})])
    #     draw(net,labels='indices')



"""
Rewired Elementary Cellular Automata
====================================

The :class:`neet.automata.reca.RewiredECA` implements a variant of an ECA
wherein the neighbors of a give cell can be specified by the user. This
allows one to study, for example, the role of topology in the dynamics of a
network. Every ``ECA`` can be represented as a ``RewiredECA`` with standard
wiring, but all ``RewiredECA`` are *fixed sized* networks.

.. rubric:: Examples

.. doctest:: automata

    >>> ca = RewiredECA(30, size=3)
    >>> ca.update([0, 1, 0])
    [1, 1, 1]
    >>> ca = RewiredECA(30, wiring=[[0,1,3], [1,1,1], [2,1,2]])
    >>> ca.update([0, 1, 0])
    [1, 0, 1]
"""
import numpy as np
from neet.statespace import StateSpace
from . import eca


class RewiredECA(eca.ECA):
    """
    RewiredECA is a class to represent elementary cellular automata rules with
    arbitrarily defined topology. Since the topology must be provided,
    RewiredECA are naturally fixed-sized.
    """

    def __init__(self, code, boundary=None, size=None, wiring=None):
        """
        Construct a rewired elementary cellular automaton rule.

        .. rubric:: Examples

        .. doctest:: automata

            >>> reca = RewiredECA(30, size=3)
            >>> reca.code
            30
            >>> reca.size
            3
            >>> reca.wiring
            array([[-1,  0,  1],
                   [ 0,  1,  2],
                   [ 1,  2,  3]])

        .. doctest:: automata

            >>> reca = RewiredECA(30, wiring=[[0,1,2],[-1,0,0],[2,3,1]])
            >>> reca.code
            30
            >>> reca.size
            3
            >>> reca.wiring
            array([[ 0,  1,  2],
                   [-1,  0,  0],
                   [ 2,  3,  1]])

        :param code: the 8-bit Wolfram code for the rule
        :type code: int
        :param boundary: the boundary conditions for the CA
        :type boundary: tuple or None
        :param size: the number of cells in the lattice
        :type size: int or None
        :param wiring: a wiring matrix
        :raises ValueError: if ``size is None and wiring is None``
        :raises ValueError: if ``size is not None and wiring is not None``
        :raises TypeError: if ``size is not None and not
                           isinstance(size, int)``
        :raises ValueError: if ``size is not None and size <= 0``
        :raises TypeError: if ``not isinstance(wiring, list) and not
                           isinstance(wiring, numpy.ndarray)``
        :raises ValueError: if ``wiring`` is not :math:`3 \times N`
        :raises ValueError: if ``any(wiring < -1) or any(wiring > N)``
        """
        super(RewiredECA, self).__init__(code, boundary=boundary)
        if size is not None and wiring is not None:
            raise ValueError("cannot provide size and wiring at the same time")
        elif size is not None:
            if not isinstance(size, int):
                raise TypeError("size must be an int")
            elif size <= 0:
                raise ValueError("size must be positive, nonzero")
            else:
                self.__size = size
                self.__wiring = np.zeros((3, size), dtype=int)
                self.__wiring[0, :] = range(-1, size - 1)
                self.__wiring[1, :] = range(0, size)
                self.__wiring[2, :] = range(1, size + 1)
        elif wiring is not None:
            if not isinstance(wiring, (list, np.ndarray)):
                raise TypeError("wiring must be a list or an array")
            wiring_array = np.copy(wiring)
            shape = wiring_array.shape
            if wiring_array.ndim != 2:
                raise ValueError("wiring must be a matrix")
            elif shape[0] != 3:
                raise ValueError("wiring must have 3 rows")
            elif np.any(wiring_array < -1):
                raise ValueError("invalid input node in wiring")
            elif np.any(wiring_array > shape[1]):
                raise ValueError("invalid input node in wiring")
            self.__size = int(shape[1])
            self.__wiring = wiring_array
        else:
            raise ValueError("either size or wiring must be provided")

    @property
    def wiring(self):
        """
        The wiring matrix for the rule.

        .. rubric:: Examples

        .. doctest:: automata

            >>> reca = RewiredECA(30, size=3)
            >>> reca.wiring
            array([[-1,  0,  1],
                   [ 0,  1,  2],
                   [ 1,  2,  3]])
            >>> eca = RewiredECA(30, wiring=[[0,1],[1,1],[-1,-1]])
            >>> eca.wiring
            array([[ 0,  1],
                   [ 1,  1],
                   [-1, -1]])

        :type: ``numpy.ndarray``
        """
        return self.__wiring

    @property
    def size(self):
        """
        The number of cells in the CA lattice.

        .. rubric:: Examples

        .. doctest:: automata

            >>> eca = RewiredECA(30, size=3)
            >>> eca.size
            3
            >>> eca = RewiredECA(30, wiring=[[-1,0], [0,1], [1,0]])
            >>> eca.size
            2

        :type: int
        """
        return self.__size

    def state_space(self):
        """
        Return a :class:`neet.statespace.StateSpace` object for the
        cellular automaton lattice.

        .. rubric:: Examples

        .. doctest:: automata

            >>> eca = RewiredECA(30, size=3)
            >>> eca.state_space()
            <neet.statespace.StateSpace object at 0x...>
            >>> space = eca.state_space()
            >>> list(space)
            [[0, 0, 0], [1, 0, 0], [0, 1, 0], [1, 1, 0], [0, 0, 1], [1, 0, 1], [0, 1, 1], [1, 1, 1]]

        :returns: :class:`neet.statespace.StateSpace`
        """
        return StateSpace(self.__size, base=2)

    def _unsafe_update(self, lattice, index=None, pin=None, values=None):
        """
        Update the state of the ``lattice``, in place, without
        checking the validity of the arguments.

        :param lattice: the one-dimensional sequence of states
        :param index: the index to update (optional)
        :param pin: a sequence of indices to fix (optional)
        :param values: a dict of index/value pairs to set (optional)
        :returns: the updated lattice
        """
        pin_states = pin is not None and pin != []
        if self.boundary:
            left = self.boundary[0]
            right = self.boundary[1]
        else:
            left = lattice[-1]
            right = lattice[0]

        code = self.code
        wiring = self.wiring
        size = len(lattice)

        if index is None:
            if pin_states:
                pinned = np.asarray(lattice)[pin]
            temp = np.copy(lattice)
            for j in range(size):
                shift = 0
                for i in range(3):
                    k = wiring[i, j]
                    if k == -1:
                        shift = 2 * shift + left
                    elif k == size:
                        shift = 2 * shift + right
                    else:
                        shift = 2 * shift + lattice[k]
                temp[j] = 1 & (code >> (7 & shift))
            lattice[:] = temp[:]
            if pin_states:
                for j, i in enumerate(pin):
                    lattice[i] = pinned[j]
        else:
            if index < 0:
                index += len(lattice)
            shift = 0
            for i in range(3):
                k = wiring[i, index]
                if k == -1:
                    shift = 2 * shift + left
                elif k == size:
                    shift = 2 * shift + right
                else:
                    shift = 2 * shift + lattice[k]
                lattice[index] = 1 & (code >> (7 & shift))

        if values is not None:
            for key in values:
                lattice[key] = values[key]

        return lattice

    def update(self, lattice, index=None, pin=None, values=None):
        """
        Update the state of the ``lattice`` in place.

        .. rubric:: Examples

        .. doctest:: automata

            >>> reca = RewiredECA(30, size=5)
            >>> reca.update([1,0,0,0,0])
            [1, 1, 0, 0, 1]
            >>> reca.wiring[:,:] = [[-1, 2, 1, 2, -1], [0, 1, 2, 3, 4], [0, 2, 3, 4, 5]]
            >>> reca.update([0,0,1,0,0])
            [0, 0, 1, 1, 0]
            >>> reca.update([1,1,1,1,1])
            [0, 0, 0, 0, 0]
            >>> reca.update([1,1,1,1,1], index=2)
            [1, 1, 0, 1, 1]
            >>> reca.update([1,1,1,1,1], pin=[1, 3])
            [0, 1, 0, 1, 0]
            >>> reca.update([1,1,1,1,1], values={0: 1, -1: 1})
            [1, 0, 0, 0, 1]

        :param lattice: the one-dimensional sequence of states
        :param index: the index to update (optional)
        :param pin: a sequence of indices to fix (optional)
        :param values: a dict of index/value pairs to set (optional)
        :returns: the updated lattice
        """
        size = len(lattice)
        if lattice not in self.state_space():
            msg = "the provided state is not in the RewiredECA's state space"
            raise ValueError(msg)

        if index is not None:
            if index < -size:
                raise IndexError("lattice index out of range")
            elif pin is not None and pin != []:
                msg = "cannot provide both the index and pin arguments"
                raise ValueError(msg)
            elif values is not None and values != {}:
                msg = "cannot provide both the index and values arguments"
                raise ValueError(msg)
        elif pin is not None and values is not None:
            for key in values.keys():
                if key in pin:
                    raise ValueError("cannot set a value for a pinned state")
        if values is not None:
            for val in values.values():
                if val != 0 and val != 1:
                    raise ValueError("invalid state in values argument")

        return self._unsafe_update(lattice, index=index, pin=pin, values=values)



import unittest
import numpy as np
from neet.automata.reca import RewiredECA


class TestRewiredECA(unittest.TestCase):
    """
    Unit tests of the RewiredECA class
    """

    def test_is_network(self):
        """
        Ensure that RewiredECA meets the requirement of a network
        """
        from neet.interfaces import is_network
        self.assertTrue(is_network(RewiredECA))
        self.assertTrue(is_network(RewiredECA(23, size=3)))
        self.assertTrue(is_network(RewiredECA(
            30, wiring=[[-1, 0, 1], [0, 1, 2], [1, 2, 3]])))

    def test_is_fixed_sized(self):
        """
        Ensure that RewiredECA is of fixed size
        """
        from neet.interfaces import is_fixed_sized
        self.assertTrue(is_fixed_sized(RewiredECA))
        self.assertTrue(is_fixed_sized(RewiredECA(23, size=3)))
        self.assertTrue(is_fixed_sized(RewiredECA(
            30, wiring=[[-1, 0, 1], [0, 1, 2], [1, 2, 3]])))

    def test_invalid_code(self):
        """
        Ensure that init fails when an invalid Wolfram code is provided
        """
        with self.assertRaises(ValueError):
            RewiredECA(-1, size=3)
        with self.assertRaises(ValueError):
            RewiredECA(256, size=3)
        with self.assertRaises(TypeError):
            RewiredECA("30", size=3)

    def test_invalid_boundary(self):
        """
        Ensure that init fails when an invalid boundary condition is provided
        """
        with self.assertRaises(TypeError):
            RewiredECA(30, boundary=[1, 2], size=3)
        with self.assertRaises(ValueError):
            RewiredECA(30, boundary=(1, 0, 1), size=3)
        with self.assertRaises(ValueError):
            RewiredECA(30, boundary=(1, 2), size=3)

    def test_invalid_size(self):
        """
        Ensure that init fails when an invalid size is provided
        """
        with self.assertRaises(TypeError):
            RewiredECA(30, size="3")
        with self.assertRaises(ValueError):
            RewiredECA(30, size=-1)
        with self.assertRaises(ValueError):
            RewiredECA(30, size=0)

    def test_invalid_wiring(self):
        """
        Ensure that init fails when an invalid wiring is provided
        """
        with self.assertRaises(TypeError):
            RewiredECA(30, wiring=5)
        with self.assertRaises(TypeError):
            RewiredECA(30, wiring="apples")
        with self.assertRaises(ValueError):
            RewiredECA(30, wiring=[])
        with self.assertRaises(ValueError):
            RewiredECA(30, wiring=[-1, 0, 1])
        with self.assertRaises(ValueError):
            RewiredECA(30, wiring=np.asarray([-1, 0, 1]))
        with self.assertRaises(ValueError):
            RewiredECA(30, wiring=[[0], [0]])
        with self.assertRaises(ValueError):
            RewiredECA(30, wiring=[[0, 0], [0], [0]])
        with self.assertRaises(ValueError):
            RewiredECA(30, wiring=[[-2], [0], [0]])
        with self.assertRaises(ValueError):
            RewiredECA(30, wiring=[[2], [0], [0]])

    def test_invalid_size_wiring(self):
        """
        Ensure that size and wiring are not both provided, but at least one is
        """
        with self.assertRaises(ValueError):
            RewiredECA(30, size=3, wiring=[])
        with self.assertRaises(ValueError):
            RewiredECA(30)
        with self.assertRaises(ValueError):
            RewiredECA(30, boundary=(0, 0))

    def test_size_init(self):
        """
        Ensure that size initialization is working correctly
        """
        eca = RewiredECA(30, size=2)
        self.assertEqual(30, eca.code)
        self.assertEqual(2, eca.size)
        self.assertTrue(np.array_equal([[-1, 0], [0, 1], [1, 2]], eca.wiring))

        eca = RewiredECA(23, boundary=(1, 0), size=5)
        self.assertEqual(23, eca.code)
        self.assertEqual(5, eca.size)
        self.assertTrue(
            np.array_equal([[-1, 0, 1, 2, 3], [0, 1, 2, 3, 4],
                            [1, 2, 3, 4, 5]],
                           eca.wiring))

    def test_wiring_init(self):
        """
        Ensure that wiring initialization is working correctly
        """
        eca = RewiredECA(30, wiring=[[0], [0], [0]])
        self.assertEqual(30, eca.code)
        self.assertEqual(1, eca.size)
        self.assertTrue(np.array_equal([[0], [0], [0]], eca.wiring))

        eca = RewiredECA(23, boundary=(1, 0), wiring=[
                         [-1, -1, 0], [0, 1, 1], [2, 0, -1]])
        self.assertEqual(23, eca.code)
        self.assertEqual(3, eca.size)
        self.assertTrue(
            np.array_equal([[-1, -1, 0], [0, 1, 1], [2, 0, -1]],
                           eca.wiring))

    def test_setting_wiring(self):
        """
        Ensure that we cannot reshape the wiring
        """
        eca = RewiredECA(30, size=2)
        with self.assertRaises(AttributeError):
            eca.wiring = np.asarray([[0, 0, 0], [0, 0, 0], [0, 0, 0]])
        self.assertEqual(-1, eca.wiring[0, 0])
        eca.wiring[0, 0] = 0
        self.assertEqual(0, eca.wiring[0, 0])

    def test_invalid_lattice_size(self):
        """
        Ensure that update fails when the lattice is the wrong size
        """
        eca = RewiredECA(30, size=3)
        with self.assertRaises(ValueError):
            eca.update([])
        with self.assertRaises(ValueError):
            eca.update([0])
        with self.assertRaises(ValueError):
            eca.update([0, 0])

    def test_invalid_lattice_state(self):
        """
        Ensure that the states of the lattice are binary states
        """
        eca = RewiredECA(30, size=3)
        with self.assertRaises(ValueError):
            eca.update([-1, 0, 1])
        with self.assertRaises(ValueError):
            eca.update([1, 0, -1])
        with self.assertRaises(ValueError):
            eca.update([2, 0, 0])
        with self.assertRaises(ValueError):
            eca.update([1, 0, 2])
        with self.assertRaises(ValueError):
            eca.update([[1], [0], [2]])
        with self.assertRaises(ValueError):
            eca.update("101")

    def test_reproduce_closed_ecas(self):
        """
        Ensure that RewiredECAs can reproduce closed ECAs
        """
        from neet.automata import ECA
        reca = RewiredECA(30, size=7)
        eca = ECA(30)
        state = [0, 0, 0, 1, 0, 0, 0]
        for _ in range(10):
            expect = eca.update(np.copy(state))
            got = reca.update(state)
            self.assertTrue(np.array_equal(expect, got))

    def test_reproduce_open_ecas(self):
        """
        Ensure that RewiredECAs can reproduce open ECAs
        """
        from neet.automata import ECA
        reca = RewiredECA(30, boundary=(1, 0), size=7)
        eca = ECA(30, boundary=(1, 0))
        state = [0, 0, 0, 1, 0, 0, 0]
        for _ in range(10):
            expect = eca.update(np.copy(state))
            got = reca.update(state)
            self.assertTrue(np.array_equal(expect, got))

    def test_rewired_network(self):
        """
        Test a non-trivially rewired network
        """
        reca = RewiredECA(30, wiring=[
            [-1, 0, 1, 2, 3], [0, 1, 2, 3, 4], [1, 2, 3, 4, 5]
        ])
        state = [0, 0, 0, 0, 1]
        self.assertEqual([1, 0, 0, 1, 1], reca.update(state))

        reca.wiring[:, :] = [
            [0, 4, 1, 2, 3], [0, 1, 2, 3, 4], [0, 2, 3, 4, 5]
        ]
        state = [0, 0, 0, 0, 1]
        self.assertEqual([0, 1, 0, 1, 1], reca.update(state))
        self.assertEqual([0, 0, 0, 1, 0], reca.update(state))
        self.assertEqual([0, 0, 1, 1, 1], reca.update(state))
        self.assertEqual([0, 0, 1, 0, 0], reca.update(state))
        self.assertEqual([0, 1, 1, 1, 0], reca.update(state))
        self.assertEqual([0, 1, 0, 0, 1], reca.update(state))
        self.assertEqual([0, 0, 1, 1, 1], reca.update(state))

    def test_reca_invalid_index(self):
        """
        Test for invalid index arguments
        """
        reca = RewiredECA(30, wiring=[
            [0, 4, 1, 2, 3], [0, 1, 2, 3, 4], [0, 2, 3, 4, 5]
        ])

        with self.assertRaises(IndexError):
            reca.update([0, 0, 0, 0, 1], index=6)

        with self.assertRaises(IndexError):
            reca.update([0, 0, 0, 0, 1], index=-6)

    def test_reca_index(self):
        """
        Test the index argument
        """
        reca = RewiredECA(30, wiring=[
            [0, 4, 1, 2, 3], [0, 1, 2, 3, 4], [0, 2, 3, 4, 5]
        ])

        state = [0, 0, 0, 0, 1]
        self.assertEqual([0, 0, 0, 1, 1], reca.update(state, index=3))
        self.assertEqual([0, 0, 1, 1, 1], reca.update(state, index=2))
        self.assertEqual([0, 0, 1, 1, 0], reca.update(state, index=-1))
        self.assertEqual([0, 1, 1, 1, 0], reca.update(state, index=1))
        self.assertEqual([0, 1, 0, 1, 0], reca.update(state, index=-3))
        self.assertEqual([0, 1, 0, 1, 0], reca.update(state, index=0))

    def test_reca_pin_none(self):
        """
        Ensure that pin behaves correctly for nil arguments
        """
        reca = RewiredECA(30, size=5)

        xs = [0, 0, 1, 0, 0]
        self.assertEqual([0, 1, 1, 1, 0], reca.update(xs, pin=None))
        self.assertEqual([1, 1, 0, 0, 1], reca.update(xs, pin=[]))

    def test_reca_pin_index_clash(self):
        """
        Ensure ValueError is raised when index and pin are provided
        """
        reca = RewiredECA(30, size=5)

        with self.assertRaises(ValueError):
            reca.update([0, 0, 0, 0, 0], index=0, pin=[1])
        with self.assertRaises(ValueError):
            reca.update([0, 0, 0, 0, 0], index=1, pin=[1])
        with self.assertRaises(ValueError):
            reca.update([0, 0, 0, 0, 0], index=1, pin=[0, 1])

    def test_reca_pin(self):
        """
        Test the pin argument
        """
        reca = RewiredECA(30, wiring=[
            [-1, 4, 1, 2, -1], [0, 1, 2, 3, 4], [0, 2, 3, 4, 5]
        ])

        xs = [0, 0, 1, 0, 0]
        self.assertEqual([0, 0, 1, 1, 0], reca.update(xs, pin=[1]))
        self.assertEqual([0, 1, 1, 1, 0], reca.update(xs, pin=[3]))
        self.assertEqual([0, 1, 1, 1, 0], reca.update(xs, pin=[3, 2]))
        self.assertEqual([0, 1, 0, 1, 0], reca.update(xs, pin=[-2]))

        reca.boundary = (1, 1)
        xs = [0, 0, 1, 0, 0]
        self.assertEqual([1, 0, 1, 0, 0], reca.update(xs, pin=[1, 3]))
        self.assertEqual([1, 1, 1, 0, 0], reca.update(xs, pin=[-2, -5]))
        self.assertEqual([1, 1, 1, 1, 0], reca.update(xs, pin=[0, 2]))

    def test_reca_values_none(self):
        """
        Ensure that values behaves correctly for nil arguments
        """
        reca = RewiredECA(30, size=5)

        xs = [0, 0, 1, 0, 0]
        self.assertEqual([0, 1, 1, 1, 0], reca.update(xs, values=None))
        self.assertEqual([1, 1, 0, 0, 1], reca.update(xs, values={}))

    def test_reca_invalid_values(self):
        """
        Ensure ValueError is raised for invalid values
        """
        reca = RewiredECA(30, size=5)

        with self.assertRaises(ValueError):
            reca.update([0, 0, 0, 0, 0], values={0: 2})

        with self.assertRaises(ValueError):
            reca.update([0, 0, 0, 0, 0], values={0: -1})

    def test_reca_values_index_clash(self):
        """
        Ensure ValueError is raised when index and values are both provided
        """
        reca = RewiredECA(30, size=5)

        with self.assertRaises(ValueError):
            reca.update([0, 0, 0, 0, 0], index=0, values={0: 1})

        with self.assertRaises(ValueError):
            reca.update([0, 0, 0, 0, 0], index=1, values={1: 0})

        with self.assertRaises(ValueError):
            reca.update([0, 0, 0, 0, 0], index=1, values={0: 0, 1: 0})

    def test_reca_values_pin_clash(self):
        """
        Ensure ValueError is raised when pin and values are both provided
        """
        reca = RewiredECA(30, size=5)

        with self.assertRaises(ValueError):
            reca.update([0, 0, 0, 0, 0], pin=[0], values={0: 1})

        with self.assertRaises(ValueError):
            reca.update([0, 0, 0, 0, 0], pin=[1], values={1: 0})

        with self.assertRaises(ValueError):
            reca.update([0, 0, 0, 0, 0], pin=[1], values={0: 0, 1: 0})

        with self.assertRaises(ValueError):
            reca.update([0, 0, 0, 0, 0], pin=[1, 0], values={0: 0, 1: 0})

    def test_reca_values(self):
        """
        Test the values argument
        """
        reca = RewiredECA(30, wiring=[
            [-1,  4,  1,  2, -1], [0,  1,  2,  3,  4], [0,  2,  3,  4,  5]
        ])

        xs = [0, 0, 1, 0, 0]
        self.assertEqual([1, 1, 1, 1, 0], reca.update(xs, values={0: 1}))
        self.assertEqual([1, 1, 0, 0, 0], reca.update(xs, values={-1: 0}))
        self.assertEqual([1, 1, 1, 1, 1], reca.update(xs, values={-2: 1}))
        self.assertEqual([1, 0, 1, 0, 0], reca.update(
            xs, values={2: 1, -5: 1}))

        reca.boundary = (1, 1)
        xs = [0, 0, 1, 0, 0]
        self.assertEqual([0, 1, 1, 1, 0], reca.update(xs, values={0: 0}))
        self.assertEqual([1, 1, 0, 0, 1], reca.update(xs, values={-1: 1}))
        self.assertEqual([0, 0, 1, 0, 0], reca.update(xs, values={-2: 0}))
        self.assertEqual([0, 1, 0, 1, 0], reca.update(
            xs, values={2: 0, -5: 0}))



"""
.. currentmodule:: neet.boolean.conv

.. testsetup:: conv

    from neet.boolean import LogicNetwork
    from neet.boolean.conv import *
    from neet.boolean.examples import s_pombe
    from neet.synchronous import transitions

Network Type Conversion
=======================

Some algorithms are more efficiently implemented on particular network type.
This module allows the user to convert between network types, though not all
conversions are possible. For example, not every
:class:`neet.boolean.LogicNetwork` can be converted to a
:class:`neet.boolean.WTNetwork`, though the reverse direction is always
possible.
"""
from .logicnetwork import LogicNetwork
from .wtnetwork import WTNetwork


def wt_to_logic(net):
    """
    Convert a :class:`neet.boolean.WTNetwork` to a
    :class:`neet.boolean.LogicNetwork`.

    .. rubric:: Examples

    .. doctest:: conv

        >>> s_pombe_logical = wt_to_logic(s_pombe)
        >>> isinstance(s_pombe_logical, LogicNetwork)
        True
        >>> transitions(s_pombe_logical) == transitions(s_pombe)
        True

    :param net: a network to convert
    :type net: :class:`neet.boolean.WTNetwork`
    :return: an equivalent :class:`neet.boolean.LogicNetwork`
    """
    if not isinstance(net, WTNetwork):
        raise TypeError("'net' must be a WTNetwork")

    truth_table = []
    for node, weights in enumerate(net.weights):
        indices = tuple([i for i, v in enumerate(weights)
                         if v != 0 or i == node])

        conditions = set()
        for dec_state in range(2**len(indices)):
            bin_state = '{0:0{1}b}'.format(dec_state, len(indices))
            prod = sum([weights[i] * int(s)
                        for i, s in zip(indices, bin_state)])
            if (prod > net.thresholds[node] or
                    (prod == net.thresholds[node] and
                        bin_state[indices.index(node)] == '1')):
                conditions.add(bin_state)

        truth_table.append((indices, conditions))

    return LogicNetwork(truth_table, net.names, reduced=True)



"""
.. currentmodule:: neet.boolean

Boolean Networks
================

The :mod:`neet.boolean` module provides network types (:class:`WTNetwork` and
:class:`LogicNetwork`) and functions for simulating Boolean network modules.

API Documentation
-----------------
"""
from .wtnetwork import WTNetwork
from .logicnetwork import LogicNetwork



"""Unit test for LogicNetwork examples"""
import unittest
import neet.boolean.examples as ex


class TestLogicExamples(unittest.TestCase):
    def test_myeloid(self):
        self.assertEqual(11, ex.myeloid.size)
        self.assertEqual(['GATA-2', 'GATA-1', 'FOG-1', 'EKLF', 'Fli-1',
                          'SCL', 'C/EBPa', 'PU.1', 'cJun', 'EgrNab',
                          'Gfi-1'], ex.myeloid.names)
        self.assertEqual(ex.myeloid.update([1, 1, 0, 0, 0, 0, 0, 0, 0,
                                            0, 0], 0),
                         [1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0])
        self.assertEqual(ex.myeloid.update([1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0]),
                         [1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0])
        self.assertEqual(ex.myeloid.update([1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0]),
                         [1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0])
        self.assertEqual(ex.myeloid.update([0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0]),
                         [0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0])

    def test_myeloid_attractors(self):
        # Assert attractors.
        attractors = ([0, 1, 1, 1, 0, 1, 0, 0, 0, 0, 0],
                      [0, 1, 1, 0, 1, 1, 0, 0, 0, 0, 0],
                      [0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0],
                      [0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1],
                      [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                      [0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0])
        for state in attractors:
            self.assertEqual(ex.myeloid.update(state[:]), state)

        # Assert non-attractors.
        for state in ex.myeloid.state_space():
            if state not in attractors:
                self.assertNotEqual(ex.myeloid.update(state[:]), state)

    def test_myeloid_logic_expressions(self):
        self.assertEqual(['GATA-2', 'GATA-1', 'FOG-1', 'EKLF', 'Fli-1', 'SCL',
                          'C/EBPa', 'PU.1', 'cJun', 'EgrNab', 'Gfi-1'],
                         ex.myeloid_from_expr.names)
        self.assertEqual(ex.myeloid_from_expr.table, ex.myeloid.table)
        # Assert attractors.
        attractors = ([0, 1, 1, 1, 0, 1, 0, 0, 0, 0, 0],
                      [0, 1, 1, 0, 1, 1, 0, 0, 0, 0, 0],
                      [0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0],
                      [0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1],
                      [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                      [0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0])
        for state in attractors:
            self.assertEqual(ex.myeloid_from_expr.update(state[:]), state)

        # Assert non-attractors.
        for state in ex.myeloid.state_space():
            if state not in attractors:
                self.assertNotEqual(
                    ex.myeloid_from_expr.update(state[:]), state)

    def test_mouse_cortical_7B(self):
        self.assertEqual(10, ex.mouse_cortical_7B.size)
        self.assertEqual(['gF', 'gE', 'gP', 'gC', 'gS', 'pF', 'pE',
                          'pP', 'pC', 'pS'], ex.mouse_cortical_7B.names)
        self.assertEqual(ex.mouse_cortical_7B.update([1, 1, 0, 0, 0, 0,
                                                      0, 0, 0, 0], 0),
                         [0, 1, 0, 0, 0, 0, 0, 0, 0, 0])
        self.assertEqual(ex.mouse_cortical_7B.update([1, 1, 0, 0, 0, 0,
                                                      0, 0, 0, 0]),
                         [0, 1, 0, 1, 1, 1, 1, 0, 0, 0])
        self.assertEqual(ex.mouse_cortical_7B.update([1, 1, 0, 0, 0, 0,
                                                      0, 0, 0, 0]),
                         [0, 1, 0, 1, 1, 1, 1, 0, 0, 0])
        self.assertEqual(ex.mouse_cortical_7B.update([0, 0, 0, 0, 0, 0,
                                                      0, 1, 1, 1]),
                         [1, 0, 0, 0, 0, 0, 0, 0, 0, 0])

    def test_mouse_cortical_7B_attractors(self):
        # Assert attractors.
        attractors = (
            # Anterior desired attractor (reachable from desired anterior
            # initial state)
            [1, 0, 1, 0, 1, 1, 0, 1, 0, 1],

            # Posterior desired attractor (unreachable from desired
            # posterior initial state)
            [0, 1, 0, 1, 0, 0, 1, 0, 1, 0]
        )

        for state in attractors:
            self.assertEqual(ex.mouse_cortical_7B.update(state[:]), state)

        # Assert non-attractors.
        for state in ex.mouse_cortical_7B.state_space():
            if state not in attractors:
                self.assertNotEqual(
                    ex.mouse_cortical_7B.update(state[:]), state)

    def test_mouse_cortical_7B_logic_expressions(self):
        self.assertEqual(['gF', 'gE', 'gP', 'gC', 'gS', 'pF', 'pE',
                         'pP', 'pC', 'pS'],
                         ex.mouse_cortical_7B_from_expr.names)
        self.assertEqual(ex.mouse_cortical_7B_from_expr.table,
                         ex.mouse_cortical_7B.table)
        # Assert attractors.
        attractors = (
            # Anterior desired attractor (reachable from desired anterior
            # initial state)
            [1, 0, 1, 0, 1, 1, 0, 1, 0, 1],

            # Posterior desired attractor (unreachable from desired
            # posterior initial state)
            [0, 1, 0, 1, 0, 0, 1, 0, 1, 0]
        )

        for state in attractors:
            self.assertEqual(
                ex.mouse_cortical_7B_from_expr.update(state[:]), state)

        # Assert non-attractors.
        for state in ex.mouse_cortical_7B.state_space():
            if state not in attractors:
                self.assertNotEqual(
                    ex.mouse_cortical_7B_from_expr.update(state[:]), state)

    def test_mouse_cortical_7C(self):
        self.assertEqual(10, ex.mouse_cortical_7C.size)
        self.assertEqual(['gF', 'gE', 'gP', 'gC', 'gS', 'pF', 'pE',
                          'pP', 'pC', 'pS'],
                         ex.mouse_cortical_7C.names)
        self.assertEqual(ex.mouse_cortical_7C.update([1, 1, 0, 0, 0, 0,
                                                      0, 0, 0, 0], 0),
                         [0, 1, 0, 0, 0, 0, 0, 0, 0, 0])
        self.assertEqual(ex.mouse_cortical_7C.update([1, 1, 0, 0, 0, 0,
                                                      0, 0, 0, 0]),
                         [0, 1, 0, 1, 1, 1, 1, 0, 0, 0])
        self.assertEqual(ex.mouse_cortical_7C.update([1, 1, 0, 0, 0, 0,
                                                      0, 0, 0, 0]),
                         [0, 1, 0, 1, 1, 1, 1, 0, 0, 0])
        self.assertEqual(ex.mouse_cortical_7C.update([0, 0, 0, 0, 0, 0,
                                                      0, 1, 1, 1]),
                         [1, 0, 0, 0, 0, 0, 0, 0, 0, 0])

    def test_mouse_cortical_7C_attractors(self):
        # Assert attractors.
        attractors = (
            # Anterior desired attractor (reachable from desired anterior
            # initial state)
            [1, 0, 1, 0, 1, 1, 0, 1, 0, 1],
            # Posterior desired attractor (unreachable from desired posterior
            # initial state)
            [0, 1, 0, 1, 0, 0, 1, 0, 1, 0]
        )

        for state in attractors:
            self.assertEqual(ex.mouse_cortical_7C.update(state[:]), state)

        # Assert non-attractors.
        for state in ex.mouse_cortical_7C.state_space():
            if state not in attractors:
                self.assertNotEqual(
                    ex.mouse_cortical_7C.update(state[:]), state)

    def test_mouse_cortical_7C_logic_expressions(self):
        self.assertEqual(['gF', 'gE', 'gP', 'gC', 'gS', 'pF', 'pE',
                          'pP', 'pC', 'pS'],
                         ex.mouse_cortical_7C_from_expr.names)
        self.assertEqual(ex.mouse_cortical_7C_from_expr.table,
                         ex.mouse_cortical_7C.table)
        # Assert attractors.
        attractors = (
             # Anterior desired attractor (reachable from desired anterior
             # initial state)
             [1, 0, 1, 0, 1, 1, 0, 1, 0, 1],
             # Posterior desired attractor (unreachable from desired
             # posterior initial state)
             [0, 1, 0, 1, 0, 0, 1, 0, 1, 0]
        )

        for state in attractors:
            self.assertEqual(
                ex.mouse_cortical_7C_from_expr.update(state[:]), state)

        # Assert non-attractors.
        for state in ex.mouse_cortical_7C.state_space():
            if state not in attractors:
                self.assertNotEqual(
                    ex.mouse_cortical_7C_from_expr.update(state[:]), state)



"""Unit test for LogicNetwork"""
import unittest
import numpy as np
from neet.python import long
from neet.boolean import LogicNetwork
from neet.exceptions import FormatError


class TestLogicNetwork(unittest.TestCase):
    def test_is_network(self):
        from neet.interfaces import is_network
        self.assertTrue(is_network(LogicNetwork([([0], {'0'})])))

    def test_is_fixed_sized(self):
        from neet.interfaces import is_fixed_sized
        self.assertTrue(is_fixed_sized(LogicNetwork([([0], {'0'})])))

    def test_init(self):
        net = LogicNetwork([((0,), {'0'})])
        self.assertEqual(1, net.size)
        self.assertEqual([(1, {0})], net._encoded_table)

        net = LogicNetwork(
            [((1,), {'0', '1'}), ((0,), {'1'})], ['A', 'B'])
        self.assertEqual(2, net.size)
        self.assertEqual(['A', 'B'], net.names)
        self.assertEqual([(2, {0, 2}), (1, {1})], net._encoded_table)

    def test_init_fail(self):
        with self.assertRaises(TypeError):
            LogicNetwork("not a list or tuple")

        with self.assertRaises(TypeError):
            LogicNetwork([((0,), {'0'})], names="A")

        with self.assertRaises(ValueError):
            LogicNetwork([((1,), {'0', '1'}), ((0,), {'1'})], names=['A'])

        with self.assertRaises(ValueError):
            LogicNetwork([((0,), {'0'})], names=["A", "B"])

        with self.assertRaises(ValueError):
            LogicNetwork([((0,),)])

        with self.assertRaises(IndexError):
            LogicNetwork([((1,), {'0'})])

        with self.assertRaises(ValueError):
            LogicNetwork([((0,), '0')])

    def test_init_long(self):
        table = [((), set()) for _ in range(65)]
        table[0] = ((np.int64(64),), set('1'))

        mask = long(2)**64

        net = LogicNetwork(table)
        self.assertEqual(net.table, table)
        self.assertEqual(net._encoded_table[0], (mask, set([mask])))

    def test_inplace_update(self):
        net = LogicNetwork([((1,), {'0', '1'}), ((0,), {'1'})])
        state = [0, 1]
        self.assertEqual(net.update(state), [1, 0])
        self.assertEqual([1, 0], state)

    def test_update(self):
        net = LogicNetwork([((0,), {'0'})])
        self.assertEqual(net.update([0], 0), [1])
        self.assertEqual(net.update([1], 0), [0])
        self.assertEqual(net.update([0]), [1])
        self.assertEqual(net.update([1]), [0])

        net = LogicNetwork([((1,), {'0', '1'}), ((0,), {'1'})])
        self.assertEqual(net.update([0, 0], 0), [1, 0])
        self.assertEqual(net.update([0, 0], 1), [0, 0])
        self.assertEqual(net.update([0, 0]), [1, 0])
        self.assertEqual(net.update([0, 1], 0), [1, 1])
        self.assertEqual(net.update([0, 1], 1), [0, 0])
        self.assertEqual(net.update([0, 1]), [1, 0])
        self.assertEqual(net.update([1, 0], 0), [1, 0])
        self.assertEqual(net.update([1, 0], 1), [1, 1])
        self.assertEqual(net.update([1, 0]), [1, 1])
        self.assertEqual(net.update([1, 1]), [1, 1])

        net = LogicNetwork([((1, 2), {'01', '10'}),
                            ((0, 2), {(0, 1), '10', (1, 1)}),
                            ((0, 1), {'11'})])
        self.assertEqual(net.update([0, 1, 0]), [1, 0, 0])
        self.assertEqual(net.update([1, 1, 1], 1), [1, 1, 1])
        self.assertEqual(net.update([0, 0, 1]), [1, 1, 0])
        self.assertEqual(net.update([0, 0, 1], pin=[1]), [1, 0, 0])
        self.assertEqual(net.update([0, 0, 1], pin=[0, 1]), [0, 0, 0])

        self.assertEqual(net.update([0, 0, 1], values={0: 0}), [0, 1, 0])
        self.assertEqual(net.update(
            [0, 0, 1], pin=[1], values={0: 0}), [0, 0, 0])

    def test_update_exceptions(self):
        net = LogicNetwork([((1, 2), {'01', '10'}),
                            ((0, 2), {(0, 1), '10', (1, 1)}),
                            ((0, 1), {'11'})])
        with self.assertRaises(ValueError):
            net.update([0, 0])

        with self.assertRaises(ValueError):
            net.update([0, 0, 1], values={0: 2})

        with self.assertRaises(ValueError):
            net.update([0, 0, 1], pin=[0], values={0: 1})

    def test_has_metadata(self):
        net = LogicNetwork([((0,), {'0'})])
        self.assertTrue(hasattr(net, 'metadata'))
        self.assertEqual(type(net.metadata), dict)

    def test_logic_simple_read(self):
        from os.path import dirname, abspath, realpath, join

        # Determine the path to the "data" directory of the neet.test module
        DATA_PATH = join(dirname(abspath(realpath(__file__))), "data")

        # Test simple network read in
        SIMPLE_TRUTH_TABLE = join(DATA_PATH, "test_simple-truth_table.txt")
        simple = LogicNetwork.read_table(SIMPLE_TRUTH_TABLE)

        self.assertEqual(simple.names, ['A', 'B', 'C', 'D'])
        self.assertEqual(simple.table, [((1, 2), set(['11', '10'])),
                                        ((0,), set(['1'])),
                                        ((1, 2, 0), set(
                                            ['010', '011', '101'])),
                                        ((3,), set(['1']))])

    def test_logic_simple_read_no_commas(self):
        from os.path import dirname, abspath, realpath, join

        # Determine the path to the "data" directory of the neet.test module
        DATA_PATH = join(dirname(abspath(realpath(__file__))), "data")

        # Test simple network read in (no commas)
        SIMPLE_TRUTH_TABLE = join(
            DATA_PATH, "test_simple_no_commas-truth_table.txt")
        simple = LogicNetwork.read_table(SIMPLE_TRUTH_TABLE)

        self.assertEqual(simple.names, ['A', 'B', 'C', 'D'])
        self.assertEqual(simple.table, [((1, 2), set(['11', '10'])),
                                        ((0,), set(['1'])),
                                        ((1, 2, 0), set(
                                            ['010', '011', '101'])),
                                        ((3,), set(['1']))])

    def test_logic_simple_read_no_header(self):
        from os.path import dirname, abspath, realpath, join

        # Determine the path to the "data" directory of the neet.test module
        DATA_PATH = join(dirname(abspath(realpath(__file__))), "data")

        # Test simple network read in (no header)
        SIMPLE_TRUTH_TABLE = join(
            DATA_PATH, "test_simple_no_header-truth_table.txt")
        with self.assertRaises(FormatError):
            LogicNetwork.read_table(SIMPLE_TRUTH_TABLE)

    def test_logic_simple_read_no_node_headers(self):
        from os.path import dirname, abspath, realpath, join

        # Determine the path to the "data" directory of the neet.test module
        DATA_PATH = join(dirname(abspath(realpath(__file__))), "data")

        # Test simple network read in (no header)
        SIMPLE_TRUTH_TABLE = join(
            DATA_PATH, "test_simple_no_node_headers-truth_table.txt")
        with self.assertRaises(FormatError):
            LogicNetwork.read_table(SIMPLE_TRUTH_TABLE)

    def test_logic_simple_read_empty(self):
        from os.path import dirname, abspath, realpath, join

        # Determine the path to the "data" directory of the neet.test module
        DATA_PATH = join(dirname(abspath(realpath(__file__))), "data")

        # Test simple network read in (no header)
        SIMPLE_TRUTH_TABLE = join(
            DATA_PATH, "test_simple_empty-truth_table.txt")
        simple = LogicNetwork.read_table(SIMPLE_TRUTH_TABLE)

        self.assertEqual(simple.names, ['A', 'B', 'C', 'D'])
        self.assertEqual(simple.table, [((0,), set(['1'])),
                                        ((1,), set(['1'])),
                                        ((2,), set(['1'])),
                                        ((3,), set(['1']))])

    def test_logic_simple_read_custom_comment(self):
        from os.path import dirname, abspath, realpath, join

        # Determine the path to the "data" directory of the neet.test module
        DATA_PATH = join(dirname(abspath(realpath(__file__))), "data")

        # Test simple network read in (no header)
        SIMPLE_TRUTH_TABLE = join(
            DATA_PATH, "test_simple_custom_comment-truth_table.txt")
        simple = LogicNetwork.read_table(SIMPLE_TRUTH_TABLE)

        self.assertEqual(simple.names, ['A', 'B', 'C', 'D'])
        self.assertEqual(simple.table, [((1, 2), set(['11', '10'])),
                                        ((0,), set(['1'])),
                                        ((1, 2, 0), set(
                                            ['010', '011', '101'])),
                                        ((3,), set(['1']))])

    def test_neighbors_in(self):

        net = LogicNetwork([((1, 2), set(['11', '10'])),
                            ((0,), set(['1'])),
                            ((0, 1, 2), set(['010', '011', '101'])),
                            ((3,), set(['1']))])

        self.assertEqual(net.neighbors_in(2), set([0, 1, 2]))

        with self.assertRaises(TypeError):
            net.neighbors_in(2.0)

        with self.assertRaises(TypeError):
            net.neighbors_in('2')

    def test_neighbors_out(self):

        net = LogicNetwork([((1, 2), set(['11', '10'])),
                            ((0,), set(['1'])),
                            ((0, 1, 2), set(['010', '011', '101'])),
                            ((3,), set(['1']))])

        self.assertEqual(net.neighbors_out(2), set([0, 2]))

    def test_neighbors_both(self):

        net = LogicNetwork([((1, 2), set(['11', '10'])),
                            ((0,), set(['1'])),
                            ((0, 1, 2), set(['010', '011', '101'])),
                            ((3,), set(['1']))])

        self.assertEqual(net.neighbors(2), set([0, 1, 2]))

    def test_node_dependency(self):
        net = LogicNetwork([((1, 2), {'11', '10'}),
                            ((0,), {'1'}),
                            ((0, 1, 2), {'010', '011', '101', '100'})])

        self.assertTrue(net.is_dependent(0, 1))
        self.assertFalse(net.is_dependent(0, 2))

        self.assertFalse(net.is_dependent(2, 2))
        self.assertTrue(net.is_dependent(2, 0))
        self.assertTrue(net.is_dependent(2, 1))

    def test_reduce_table(self):
        table = [((1, 2), {'11', '10'}),
                 ((0,), {'1'}),
                 ((0, 1, 2), {'010', '011', '101', '100'})]
        net = LogicNetwork(table, reduced=True)

        reduced_table = [((1,), {'1'}),
                         ((0,), {'1'}),
                         ((0, 1), {'01', '10'})]
        self.assertEqual(net.table, reduced_table)

        net = LogicNetwork(table)
        self.assertEqual(net.table, table)
        self.assertEqual(net._encoded_table,
                         [(6, {6, 2}), (1, {1}), (7, {2, 6, 5, 1})])

        net.reduce_table()
        self.assertEqual(net.table, reduced_table)
        self.assertEqual(net._encoded_table,
                         [(2, {2}), (1, {1}), (3, {2, 1})])

        net.reduce_table()
        self.assertEqual(net.table,
                         [((1,), {'1'}),
                          ((0,), {'1'}),
                          ((0, 1), {'01', '10'})])

        net = LogicNetwork([((0, 1), {'00', '01', '10', '11'}),
                            ((1,), {'1'})],
                           reduced=True)

        self.assertEqual(net.table,
                         [((0,), {'0', '1'}),
                          ((1,), {'1'})])

        net = LogicNetwork([((1,), {'0', '1'}),
                            ((1,), {'1'})],
                           reduced=True)

        self.assertEqual(net.table,
                         [((0,), {'1'}),
                          ((1,), {'1'})])

    def test_to_networkx_graph_names(self):
        net = LogicNetwork([((1, 2), {'01', '10'}),
                            ((0, 2), ((0, 1), '10', [1, 1])),
                            ((0, 1), {'11'})], ['A', 'B', 'C'])

        nx_net = net.to_networkx_graph(labels='names')
        self.assertEqual(set(nx_net), set(['A', 'B', 'C']))

    def test_to_networkx_graph_names_fail(self):
        net = LogicNetwork([((0,), {'0'})])

        with self.assertRaises(ValueError):
            net.to_networkx_graph(labels='names')

    def test_to_networkx_metadata(self):
        net = LogicNetwork([((0,), {'0'})])
        net.metadata['name'] = 'net_name'

        nx_net = net.to_networkx_graph(labels='indices')

        self.assertEqual(nx_net.graph['name'], net.metadata['name'])

    # def test_draw(self):
    #     net = bnet.LogicNetwork([((0,), {'0'})])
    #     draw(net,labels='indices')



import unittest
import neet.boolean.examples as ex


def all_example_networks():
    return [
        ex.s_pombe,
        ex.s_cerevisiae,
        ex.c_elegans,
        ex.p53_no_dmg,
        ex.p53_dmg,
        ex.mouse_cortical_7B,
        ex.mouse_cortical_7C,
        ex.myeloid,
    ]


class TestBooleanExamples(unittest.TestCase):

    def test_examples_loaded(self):
        """
        Test that all example networks successfully load.
        """
        all_example_networks()

    def test_s_pombe(self):
        self.assertEqual(9, ex.s_pombe.size)
        self.assertEqual(["SK", "Cdc2_Cdc13", "Ste9", "Rum1", "Slp1",
                          "Cdc2_Cdc13_active", "Wee1_Mik1", "Cdc25",
                          "PP"], ex.s_pombe.names)

    def test_s_cerevisiae(self):
        self.assertEqual(11, ex.s_cerevisiae.size)
        self.assertEqual(["Cln3", "MBF", "SBF", "Cln1_2", "Cdh1", "Swi5",
                          "Cdc20_Cdc14", "Clb5_6", "Sic1", "Clb1_2",
                          "Mcm1_SFF"], ex.s_cerevisiae.names)

    def test_examples_metadata(self):
        """
        Test that all examples have name, description, and citation metadata.
        """
        for net in all_example_networks():
            self.assertTrue('name' in net.metadata)
            self.assertTrue('description' in net.metadata)
            self.assertTrue('citation' in net.metadata)



"""
.. currentmodule:: neet.statespace

.. testsetup:: statespace

    from neet.statespace import StateSpace

State Space
===========

A fundamental property of networks is their state space. The
:class:`StateSpace` class provides a type for representing such a state
space. It has anumber of convenient properties, in particular

1. it is iterable, :meth:`StateSpace.__iter__`
2. you can ask if it contains a particular state,
   :meth:`StateSpace.__contains__`
3. it can :meth:`StateSpace.encode` states as integers, and
   :meth:`StateSpace.decode` integers into states

All of this together facilitates the clean implementation of many of the core
functions of :mod:`neet`.

API Documentation
-----------------
"""
from .python import long


class StateSpace(object):
    """
    :class:`StateSpace` represents the state space of a network model. It may be
    either uniform, i.e. all nodes have the same base, or non-uniform.
    """

    def __init__(self, spec, base=None):
        """
        Initialize the state spec in accordance with the provided ``spec``
        and base ``base``.

        .. rubric:: Examples of Uniform State Spaces:

        .. doctest:: statespace

            >>> spec = StateSpace(5)
            >>> (spec.is_uniform, spec.ndim, spec.base)
            (True, 5, 2)
            >>> spec = StateSpace(3, base=3)
            >>> (spec.is_uniform, spec.ndim, spec.base)
            (True, 3, 3)
            >>> spec = StateSpace([2, 2, 2])
            >>> (spec.is_uniform, spec.ndim, spec.base)
            (True, 3, 2)

        .. rubric:: Examples of Non-Uniform State Spaces:

        .. doctest:: statespace

            >>> spec = StateSpace([2, 3, 4])
            >>> (spec.is_uniform, spec.base, spec.ndim)
            (False, [2, 3, 4], 3)

        :param spec: the number of nodes or an array of node bases
        :type spec: int or list
        :param base: the base of the network nodes (ignored if ``spec`` is
                     a list)
        :raises TypeError: if ``spec`` is neither an int nor a list of ints
        :raises TypeError: if ``base`` is neither ``None`` nor an int
        :raises ValueError: if ``base`` is negative or zero
        :raises ValueError: if any element of ``spec`` is negative or zero
        :raises ValueError: if ``spec`` is empty
        """
        if isinstance(spec, int):
            if spec < 1:
                raise ValueError("ndim cannot be zero or negative")
            if base is None:
                base = 2
            elif not isinstance(base, int):
                raise TypeError("base must be an int")
            elif base < 1:
                raise ValueError("base must be positive, nonzero")

            self.__is_uniform = True
            self.__ndim = spec
            self.__base = base
            self.__volume = base**spec

        elif isinstance(spec, list):
            if len(spec) == 0:
                raise ValueError("bases cannot be an empty")
            else:
                self.__is_uniform = True
                self.__volume = 1
                first_base = spec[0]
                if base is not None and first_base != base:
                    raise ValueError("base does not match base of spec")
                for spec_base in spec:
                    if not isinstance(spec_base, int):
                        raise TypeError("spec must be a list of ints")
                    elif spec_base < 1:
                        msg = "spec may only contain positive elements"
                        raise ValueError(msg)
                    if self.__is_uniform and spec_base != first_base:
                        self.__is_uniform = False
                        if base is not None:
                            raise ValueError("b does not match base of spec")
                    self.__volume *= spec_base
                self.__ndim = len(spec)
                if self.__is_uniform:
                    self.__base = first_base
                else:
                    self.__base = spec[:]
        else:
            raise TypeError("spec must be an int or a list")

    @property
    def ndim(self):
        """
        Get the dimensionality of the state space.
        """
        return self.__ndim

    @property
    def base(self):
        """
        Get the base of each direction of the state space.

        If the state space is not uniform, the result is a list of bases
        (one for each dimension). Otherwise, the result is an integer.

        :return: a list of bases, one for each dimension
        """
        return self.__base

    @property
    def volume(self):
        """
        Get the volume of the state space.
        """
        return self.__volume

    @property
    def is_uniform(self):
        """
        Get whether every direction in the state space has the same base.
        """
        return self.__is_uniform

    def __iter__(self):
        """
        Iterate over the states in the state space

        .. rubric:: Examples of Boolean Spaces

        .. doctest:: statespace

            >>> list(StateSpace(1))
            [[0], [1]]
            >>> list(StateSpace(2))
            [[0, 0], [1, 0], [0, 1], [1, 1]]
            >>> list(StateSpace(3))
            [[0, 0, 0], [1, 0, 0], [0, 1, 0], [1, 1, 0], [0, 0, 1], [1, 0, 1], [0, 1, 1], [1, 1, 1]]

        .. rubric:: Examples of Non-Boolean Spaces

        .. doctest:: statespace

            >>> list(StateSpace(1, base=3))
            [[0], [1], [2]]
            >>> list(StateSpace(2, base=4))
            [[0, 0], [1, 0], [2, 0], [3, 0], [0, 1], [1, 1], [2, 1], [3, 1], [0, 2], [1, 2], [2, 2], [3, 2], [0, 3], [1, 3], [2, 3], [3, 3]]

        .. rubric:: Examples of Non-Uniform Spaces

        .. doctest:: statespace

            >>> list(StateSpace([1,2,3]))
            [[0, 0, 0], [0, 1, 0], [0, 0, 1], [0, 1, 1], [0, 0, 2], [0, 1, 2]]
            >>> list(StateSpace([3,4]))
            [[0, 0], [1, 0], [2, 0], [0, 1], [1, 1], [2, 1], [0, 2], [1, 2], [2, 2], [0, 3], [1, 3], [2, 3]]

        :yields: each possible state in the state space
        """
        state = [0] * self.ndim
        yield state[:]
        i = 0
        while i != self.ndim:
            base = self.__base if self.__is_uniform else self.__base[i]
            if state[i] + 1 < base:
                state[i] += 1
                for j in range(i):
                    state[j] = 0
                i = 0
                yield state[:]
            else:
                i += 1

    def __contains__(self, states):
        """
        Determine if a state is in the state space

        .. rubric:: Examples:

        .. doctest:: statespace

            >>> state_space = StateSpace(3)
            >>> [0, 0, 0] in state_space
            True
            >>> [0, 0] in state_space
            False
            >>> [1, 2, 1] in state_space
            False

        .. doctest:: statespace

            >>> [0, 2, 1] in StateSpace([2, 3, 2])
            True
            >>> [0, 1] in StateSpace([2, 2, 3])
            False
            >>> [1, 1, 6] in StateSpace([2, 3, 4])
            False

        :param states: the one-dimensional sequence of node states
        :returns: ``True`` if the ``states`` are valid
        """
        try:
            if len(states) != self.ndim:
                return False

            if self.is_uniform:
                for state in states:
                    if state not in range(self.base):
                        return False
            else:
                for state, base in zip(states, self.base):
                    if state not in range(base):
                        return False
        except TypeError:
            return False
        except IndexError:
            return False

        return True

    def _unsafe_encode(self, state):
        """
        Encode a state as an integer consistent with the state space, without
        checking the validity of the arguments. The encoding is such that,
        for example, the state [1, 0, 0] will correspond to 1; the state
        [1, 1, 0] will correspond to 3.

        .. rubric:: Examples:

        .. doctest:: statespace

            >>> space = StateSpace(3, base=2)
            >>> [space._unsafe_encode(s) for s in space]
            [0, 1, 2, 3, 4, 5, 6, 7]

        .. doctest:: statespace

            >>> space = StateSpace([2,3,4])
            >>> [space._unsafe_encode(s) for s in space]
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23]


        :param state: the state to encode
        :type state: list
        :returns: a unique integer encoding of the state
        :raises ValueError: if ``state`` has an incorrect length
        """
        encoded, place = long(0), long(1)

        base = self.__base
        if self.is_uniform:
            for x in state:
                encoded += place * x
                place *= base
        else:
            for (x, b) in zip(state, base):
                encoded += place * x
                place *= b

        return long(encoded)

    def encode(self, state):
        """
        Encode a state as an integer consistent with the state space. The
        encoding is such that, for example, the state [1, 0, 0] will
        correspond to 1; the state [1, 1, 0] will correspond to 3.

        .. rubric:: Examples:

        .. doctest:: statespace

            >>> space = StateSpace(3, base=2)
            >>> [space.encode(s) for s in space]
            [0, 1, 2, 3, 4, 5, 6, 7]

        .. doctest:: statespace

            >>> space = StateSpace([2,3,4])
            >>> [space.encode(s) for s in space]
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23]


        :param state: the state to encode
        :type state: list
        :returns: a unique integer encoding of the state
        :raises ValueError: if ``state`` has an incorrect length
        """
        if state not in self:
            raise ValueError("state is not in state space")

        return self._unsafe_encode(state)

    def decode(self, encoded):
        """
        Decode an integer into a state in accordance with the state space.

        .. rubric:: Examples:

        .. doctest:: statespace

            >>> space = StateSpace(3)
            >>> list(space)
            [[0, 0, 0], [1, 0, 0], [0, 1, 0], [1, 1, 0], [0, 0, 1], [1, 0, 1], [0, 1, 1], [1, 1, 1]]
            >>> [space.decode(x) for x in range(space.volume)]
            [[0, 0, 0], [1, 0, 0], [0, 1, 0], [1, 1, 0], [0, 0, 1], [1, 0, 1], [0, 1, 1], [1, 1, 1]]

        .. doctest:: statespace

            >>> space = StateSpace([2,3])
            >>> list(space)
            [[0, 0], [1, 0], [0, 1], [1, 1], [0, 2], [1, 2]]
            >>> [space.decode(x) for x in range(space.volume)]
            [[0, 0], [1, 0], [0, 1], [1, 1], [0, 2], [1, 2]]

        :param encoded: the encoded state
        :type encoded: int
        :returns: the decoded state as a list
        """
        state = [0] * self.__ndim
        base = self.__base
        if self.is_uniform:
            b = base
            for i in range(self.__ndim):
                state[i] = encoded % b
                encoded = int(encoded / b)
        else:
            for i in range(self.__ndim):
                b = base[i]
                state[i] = encoded % b
                encoded = int(encoded / b)
        return state



import unittest
from collections import Counter
from neet.automata import ECA
from neet.boolean import WTNetwork, LogicNetwork
from neet.boolean.examples import s_pombe, s_cerevisiae, c_elegans
from neet.synchronous import (trajectory, transitions, transition_graph,
                              attractors, basins, basin_entropy,
                              timeseries, Landscape)
from neet.statespace import StateSpace
import numpy as np
import networkx as nx
from .mock import MockObject, MockNetwork, MockFixedSizedNetwork


class TestSynchronous(unittest.TestCase):
    """
    Unit tests for the ``neet.synchronous`` module
    """

    def test_trajectory_not_network(self):
        """
        ``trajectory`` should raise a type error if ``net`` is not a network
        """
        with self.assertRaises(TypeError):
            trajectory(5, [1, 2, 3])

        with self.assertRaises(TypeError):
            trajectory(MockObject(), [1, 2, 3])

        with self.assertRaises(TypeError):
            trajectory(MockFixedSizedNetwork, [1, 2, 3])

    def test_trajectory_too_short(self):
        """
        ``trajectory`` should raise a value error if ``timeseries`` is less
        than 1
        """
        with self.assertRaises(ValueError):
            trajectory(MockFixedSizedNetwork(), [1, 2, 3], timesteps=0)

        with self.assertRaises(ValueError):
            trajectory(MockFixedSizedNetwork(), [1, 2, 3], timesteps=-1)

    def test_trajectory_eca(self):
        """
        test ``trajectory`` on ECAs
        """
        rule30 = ECA(30)
        with self.assertRaises(ValueError):
            trajectory(rule30, [])

        xs = [0, 1, 0]
        got = trajectory(rule30, xs)
        self.assertEqual([0, 1, 0], xs)
        self.assertEqual([[0, 1, 0], [1, 1, 1]], got)

        got = trajectory(rule30, xs, timesteps=2)
        self.assertEqual([0, 1, 0], xs)
        self.assertEqual([[0, 1, 0], [1, 1, 1], [0, 0, 0]], got)

    def test_trajectory_eca_encoded(self):
        """
        test ``trajectory`` on ECAs; encoding the states
        """
        rule30 = ECA(30)
        with self.assertRaises(ValueError):
            trajectory(rule30, [], encode=True)

        state = [0, 1, 0]
        got = trajectory(rule30, state, encode=True)
        self.assertEqual([0, 1, 0], state)
        self.assertEqual([2, 7], got)

        got = trajectory(rule30, state, timesteps=2, encode=True)
        self.assertEqual([0, 1, 0], state)
        self.assertEqual([2, 7, 0], got)

    def test_trajectory_wtnetwork(self):
        """
        test ``trajectory`` on WTNetworks
        """
        net = WTNetwork(
            weights=[[1, 0], [-1, 0]],
            thresholds=[0.5, 0.0],
            theta=WTNetwork.positive_threshold
        )

        state = [0, 0]
        got = trajectory(net, state)
        self.assertEqual([0, 0], state)
        self.assertEqual([[0, 0], [0, 1]], got)

        got = trajectory(net, state, timesteps=3)
        self.assertEqual([0, 0], state)
        self.assertEqual([[0, 0], [0, 1], [0, 1], [0, 1]], got)

    def test_trajectory_wtnetwork_encoded(self):
        """
        test ``trajectory`` on WTNetworks; encoding the states
        """
        net = WTNetwork(
            weights=[[1, 0], [-1, 0]],
            thresholds=[0.5, 0.0],
            theta=WTNetwork.positive_threshold
        )

        state = [0, 0]
        got = trajectory(net, state, encode=True)
        self.assertEqual([0, 0], state)
        self.assertEqual([0, 2], got)

        got = trajectory(net, state, timesteps=3, encode=True)
        self.assertEqual([0, 0], state)
        self.assertEqual([0, 2, 2, 2], got)

    def test_trajectory_logicnetwork(self):
        """
        test `trajectory` on `LogicNetwork`s
        """
        net = LogicNetwork([((1, 2), {'01', '10'}),
                            ((0, 2), {'01', '10', '11'}),
                            ((0, 1), {'11'})])
        state = [0, 1, 0]
        got = trajectory(net, state, 3)
        self.assertEqual([[0, 1, 0], [1, 0, 0], [0, 1, 0], [1, 0, 0]], got)
        self.assertEqual([0, 1, 0], state)

        got = list(trajectory(net, state, 3, encode=True))
        self.assertEqual([2, 1, 2, 1], got)

    def test_transitions_not_network(self):
        """
        ``transitions`` should raise a type error if ``net`` is not a network
        """
        with self.assertRaises(TypeError):
            transitions(MockObject(), 5)

    def test_transitions_not_fixed_sized(self):
        """
        ``transitions`` should raise an error if ``net`` is not fixed sized
        and ``size`` is ``None``
        """
        with self.assertRaises(ValueError):
            transitions(ECA(30), size=None)

    def test_transitions_fixed_sized(self):
        """
        ``transitions`` should raise an error if ``net`` is fixed sized and
        ``size`` is not ``None``
        """
        with self.assertRaises(ValueError):
            transitions(MockFixedSizedNetwork, size=3)

    def test_transitions_eca(self):
        """
        test ``transitions`` on ECAs; encoding the states
        """
        rule30 = ECA(30)

        got = transitions(rule30, size=1)
        self.assertEqual([[0], [0]], got)

        got = transitions(rule30, size=2)
        self.assertEqual([[0, 0], [1, 0], [0, 1], [0, 0]], got)

        got = transitions(rule30, size=3)
        self.assertEqual([[0, 0, 0], [1, 1, 1], [1, 1, 1], [1, 0, 0],
                          [1, 1, 1], [0, 0, 1], [0, 1, 0], [0, 0, 0]], got)

    def test_transitions_eca_encoded(self):
        """
        test ``transitions`` on ECAs; encoding the states
        """
        rule30 = ECA(30)

        got = transitions(rule30, size=1, encode=True)
        self.assertEqual([0, 0], got)

        got = transitions(rule30, size=2, encode=True)
        self.assertEqual([0, 1, 2, 0], got)

        got = transitions(rule30, size=3, encode=True)
        self.assertEqual([0, 7, 7, 1, 7, 4, 2, 0], got)

    def test_transitions_wtnetwork(self):
        """
        test ``transitions`` on WTNetworks
        """
        net = WTNetwork(
            weights=[[1, 0], [-1, 1]],
            thresholds=[0.5, 0.0],
            theta=WTNetwork.positive_threshold
        )

        got = transitions(net)
        self.assertEqual([[0, 1], [1, 0], [0, 1], [1, 1]], got)

    def test_transitions_wtnetwork_encoded(self):
        """
        test ``transitions`` on WTNetworks; encoding the states
        """
        net = WTNetwork(
            weights=[[1, 0], [-1, 1]],
            thresholds=[0.5, 0.0],
            theta=WTNetwork.positive_threshold
        )

        got = transitions(net, encode=True)
        self.assertEqual([2, 1, 2, 3], got)

    def test_transitions_logicnetwork(self):
        """
        test `transitions` on `LogicNetwork`s
        """
        net = LogicNetwork([((1,), {'0', '1'}), ((0,), {'1'})])
        got = transitions(net)
        self.assertEqual([[1, 0], [1, 1], [1, 0], [1, 1]], got)

    def test_transitions_logicnetwork_encoded(self):
        """
        test `transitions` on `LogicNetwork`s, states encoded
        """
        net = LogicNetwork([((1,), {'0', '1'}), ((0,), {'1'})])
        got = transitions(net, encode=True)
        self.assertEqual([1, 3, 1, 3], got)

    def test_transition_graph_not_network(self):
        """
        ``transitions_graph`` should raise an error if ``net`` is not a network
        """
        with self.assertRaises(TypeError):
            transition_graph(MockObject())

    def test_transition_graph_variable_sized(self):
        """
        ``transitions_graph`` should raise an error if ``net`` is variable
        sized and ``size`` is ``None``
        """
        with self.assertRaises(ValueError):
            transition_graph(ECA(30))

    def test_transition_graph_fixed_sized(self):
        """
        ``transitions_graph`` should raise an error if ``net`` is fixed sized
        and ``size`` is not ``None``
        """
        with self.assertRaises(ValueError):
            transition_graph(MockFixedSizedNetwork(), size=5)

    def test_transition_graph_eca(self):
        """
        test ``transitions_graph`` on ``ECA``
        """
        graph = transition_graph(ECA(30), size=8)
        self.assertEqual(256, graph.number_of_nodes())
        self.assertEqual(256, graph.number_of_edges())

    def test_transition_graph_s_pombe(self):
        """
        test ``transitions_graph`` on ``s_pombe``
        """
        volume = s_pombe.state_space().volume
        graph = transition_graph(s_pombe)
        self.assertEqual(volume, graph.number_of_nodes())
        self.assertEqual(volume, graph.number_of_edges())

    def test_attractors_invalid_net(self):
        """
        ``attractors`` should raise an error if ``net`` is neither a network
        nor a networkx digraph
        """
        with self.assertRaises(TypeError):
            attractors('blah')

        with self.assertRaises(TypeError):
            attractors(MockObject())

        with self.assertRaises(TypeError):
            attractors(nx.Graph())

    def test_attractors_variable_sized(self):
        """
        ``attractors`` should raise an error if ``net`` is a variable sized
        network and ``size`` is ``None``
        """
        with self.assertRaises(ValueError):
            attractors(ECA(30), size=None)

    def test_attractors_fixed_sized(self):
        """
        ``attractors`` should raise an error if ``net`` is either a fixed sized
        network or a networkx digraph, and ``size`` is not ``None``
        """
        with self.assertRaises(ValueError):
            attractors(MockFixedSizedNetwork(), size=5)

        #  with self.assertRaises(ValueError):
        #      attractors(nx.DiGraph(), size=5)

    def test_attractors_eca(self):
        """
        test ``attractors`` on ECA
        """
        networks = [(ECA(30), 2, 3), (ECA(30), 3, 1), (ECA(30), 4, 4),
                    (ECA(30), 5, 2), (ECA(30), 6, 3), (ECA(110), 2, 1),
                    (ECA(110), 3, 1), (ECA(110), 4, 3), (ECA(110), 5, 1),
                    (ECA(110), 6, 3)]
        for rule, width, size in networks:
            self.assertEqual(size, len(attractors(rule, width)))

    def test_attractors_wtnetworks(self):
        """
        test ``attractors`` on WTNetworks
        """
        networks = [(s_pombe, 13), (s_cerevisiae, 7), (c_elegans, 5)]
        for net, size in networks:
            self.assertEqual(size, len(attractors(net)))

    def test_attractors_transition_graph(self):
        """
        test ``attractors`` on ``s_pombe`` transition graph
        """
        att_from_graph = attractors(transition_graph(s_pombe))
        att_from_network = attractors(s_pombe)

        for (a, b) in zip(att_from_graph, att_from_network):
            a.sort()
            b.sort()

        att_from_graph.sort()
        att_from_network.sort()

        self.assertEqual(att_from_network, att_from_graph)

    def test_basins_invalid_net(self):
        """
        ``basins`` should raise an error if ``net`` is neither a network nor a
        networkx digraph
        """
        with self.assertRaises(TypeError):
            basins('blah')

        with self.assertRaises(TypeError):
            basins(MockObject)

        with self.assertRaises(TypeError):
            basins(nx.Graph())

    def test_basins_variable_sized(self):
        """
        ``basins`` should raise an error if ``net`` is a variable sized network
        and ``size`` is ``None``
        """
        with self.assertRaises(ValueError):
            basins(ECA(30), size=None)

    def test_basins_fixed_sized(self):
        """
        ``basins`` should raise an error if ``net`` is a fized sized network
        and ``size`` is not ``None``
        """
        with self.assertRaises(ValueError):
            basins(MockFixedSizedNetwork, size=5)

    def test_basins_transition_graph(self):
        """
        test ``basins`` on ``s_pombe`` transition graph
        """
        from_graph = basins(transition_graph(s_pombe))
        from_network = basins(s_pombe)

        edges_from_graph = [g.edges() for g in from_graph]
        edges_from_network = [g.edges() for g in from_network]

        self.assertEqual(edges_from_network, edges_from_graph)

    def test_basins_eca(self):
        """
        test ``basins`` on ECAs
        """
        networks = [(ECA(30), 2, [2, 1, 1]), (ECA(30), 3, [8]),
                    (ECA(30), 4, [2, 12, 1, 1]), (ECA(30), 5, [2, 30]),
                    (ECA(30), 6, [62, 1, 1]), (ECA(110), 2, [4]),
                    (ECA(110), 3, [8]), (ECA(110), 4, [4, 6, 6]),
                    (ECA(110), 5, [32]), (ECA(110), 6, [10, 27, 27])]

        for net, width, basin_sizes in networks:
            basin_counter = Counter([len(c) for c in basins(net, size=width)])
            self.assertEqual(Counter(basin_sizes), basin_counter)

    def test_basins_wtnetwork(self):
        """
        test ``basins`` on WTNetworks
        """
        networks = [(s_pombe, [378, 2, 2, 2, 104, 6, 6, 2, 2, 2, 2, 2, 2]),
                    (s_cerevisiae, [7, 1764, 151, 1, 9, 109, 7]),
                    (c_elegans, [4, 219, 12, 5, 16])]

        for net, basin_sizes in networks:
            basin_counter = Counter([len(c) for c in basins(net)])
            self.assertEqual(Counter(basin_sizes), basin_counter)

    def test_basin_entropy_invalid_net(self):
        """
        ``basin_entropy`` should raise an error if ``net`` is neither
        a network nor a networkx digraph
        """
        with self.assertRaises(TypeError):
            basin_entropy('blee')

        with self.assertRaises(TypeError):
            basin_entropy(MockObject)

        with self.assertRaises(TypeError):
            basin_entropy(nx.Graph())

    def test_basin_entropy_variable_sized(self):
        """
        ``basin_entropy`` should raise an error if ``net`` is a variable
        sized network and ``size`` is ``None``
        """
        with self.assertRaises(ValueError):
            basin_entropy(ECA(30), size=None)

    def test_basin_entropy_fixed_sized(self):
        """
        ``basin_entropy`` should raise an error if ``net`` is a fized
        sized network and ``size`` is not ``None``
        """
        with self.assertRaises(ValueError):
            basin_entropy(MockFixedSizedNetwork, size=5)

    def test_basin_entropy_transition_graph(self):
        """
        test ``basin_entropy`` on ``s_pombe`` transition graph
        """
        from_graph = basin_entropy(transition_graph(s_pombe))
        from_network = basin_entropy(s_pombe)

        self.assertAlmostEqual(from_network, from_graph)

    def test_basin_entropy_eca(self):
        """
        test ``basin_entropy`` on ECAs
        """
        networks = [(ECA(30), 2, 1.5), (ECA(30), 3, 0.),
                    (ECA(30), 4, 1.186278124459133),
                    (ECA(30), 5, 0.3372900666170139),
                    (ECA(30), 6, 0.23187232431271465),
                    (ECA(110), 2, 0.), (ECA(110), 3, 0.),
                    (ECA(110), 4, 1.561278124459133),
                    (ECA(110), 5, 0.),
                    (ECA(110), 6, 1.4690124052234232)]

        for net, width, entropy in networks:
            self.assertAlmostEqual(basin_entropy(net, size=width), entropy)

    def test_basin_entropy_wtnetwork(self):
        """
        test ``basin_entropy`` on WTNetworks
        """
        networks = [(s_pombe, 1.2218888338849747),
                    (s_cerevisiae, 0.7838577302128783),
                    (c_elegans, 0.8542673572822357),
                    ]

        for net, entropy in networks:
            self.assertAlmostEqual(basin_entropy(net), entropy)

    def test_basin_entropy_wtnetwork_base10(self):
        """
        test ``basin_entropy`` on WTNetworks with different base
        """
        networks = [(s_pombe, 0.36782519036626099),
                    (s_cerevisiae, 0.2359646891271609),
                    (c_elegans, 0.2571600988585521),
                    ]

        for net, entropy in networks:
            self.assertAlmostEqual(basin_entropy(net, base=10), entropy)

    def test_timeseries_not_network(self):
        """
        ``timeseries`` should raise an error if ``net`` is not a network
        """
        with self.assertRaises(TypeError):
            timeseries(5, timesteps=2)

        with self.assertRaises(TypeError):
            timeseries(MockObject(), timesteps=2)

    def test_timeseries_variable_sized(self):
        """
        ``timeseries`` should raise an error if ``net`` is variable sized and
        ``size`` is ``None``
        """
        with self.assertRaises(ValueError):
            timeseries(ECA(30), size=None, timesteps=5)

    def test_timeseries_fixed_sized(self):
        """
        ``timeseries`` should raise an error if ``net`` is fixed sized and
        ``size`` is not ``None``
        """
        with self.assertRaises(ValueError):
            timeseries(MockFixedSizedNetwork, size=5, timesteps=5)

    def test_timeseries_too_short(self):
        """
        ``timeseries`` shoudl raise an error if ``timesteps`` is too small
        """
        with self.assertRaises(ValueError):
            timeseries(MockFixedSizedNetwork(), timesteps=0)

        with self.assertRaises(ValueError):
            timeseries(MockFixedSizedNetwork(), timesteps=-1)

    def test_timeseries_eca(self):
        """
        test ``timeseries`` on ECA
        """
        rule = ECA(30)
        for size in [5, 7, 11]:
            time = 10
            series = timeseries(rule, timesteps=time, size=size)
            self.assertEqual((size, 2**size, time + 1), series.shape)
            for index, state in enumerate(rule.state_space(size)):
                traj = trajectory(rule, state, timesteps=time)
                for t, expect in enumerate(traj):
                    got = series[:, index, t]
                    self.assertTrue(np.array_equal(expect, got))

    def test_timeseries_wtnetworks(self):
        """
        test ``timeseries`` on WTNetwork
        """
        for (net, size) in [(s_pombe, 9), (s_cerevisiae, 11), (c_elegans, 8)]:
            time = 10
            series = timeseries(net, timesteps=time)
            self.assertEqual((size, 2**size, time + 1), series.shape)
            for index, state in enumerate(net.state_space()):
                traj = trajectory(net, state, timesteps=time)
                for t, expect in enumerate(traj):
                    got = series[:, index, t]
                    self.assertTrue(np.array_equal(expect, got))


class TestLandscape(unittest.TestCase):
    """
    Test the ``neet.synchronous.Landscape`` class
    """

    def test_canary(self):
        """
        Canary test
        """
        self.assertTrue(True)

    def test_init_not_network(self):
        """
        ``Landscape.__init__`` should raise a type error if ``net`` is not a
        network.
        """
        with self.assertRaises(TypeError):
            Landscape(MockObject())

        with self.assertRaises(TypeError):
            Landscape(MockObject(), size=5)

    def test_init_not_fixed_sized(self):
        """
        ``Landscape.__init__`` should raise a value error if ``net`` is not
        fixed sized and ``size`` is ``None``.
        """
        with self.assertRaises(ValueError):
            Landscape(MockNetwork())

        with self.assertRaises(ValueError):
            Landscape(MockNetwork(), size=None)

    def test_init_fixed_sized(self):
        """
        ``Landscape.__init__`` should raise a value errorif ``net`` is fixed
        sized, but ``size`` is not ``None``.
        """
        with self.assertRaises(ValueError):
            Landscape(MockFixedSizedNetwork(), size=3)

    def test_transitions_eca(self):
        ca = ECA(30)

        landscape = Landscape(ca, size=1)
        self.assertEqual(ca, landscape.network)
        self.assertEqual(1, landscape.size)
        self.assertEqual([0, 0], list(landscape.transitions))

        landscape = Landscape(ca, size=2)
        self.assertEqual(ca, landscape.network)
        self.assertEqual(2, landscape.size)
        self.assertEqual([0, 1, 2, 0], list(landscape.transitions))

        landscape = Landscape(ca, size=3)
        self.assertEqual(ca, landscape.network)
        self.assertEqual(3, landscape.size)
        self.assertEqual([0, 7, 7, 1, 7, 4, 2, 0], list(landscape.transitions))

        landscape = Landscape(ca, size=10)
        trans = landscape.transitions
        self.assertEqual(ca, landscape.network)
        self.assertEqual(10, landscape.size)
        self.assertEqual(1024, len(trans))
        self.assertEqual([0, 515, 7, 517, 14, 525, 11,
                          521, 28, 543], list(trans[:10]))
        self.assertEqual([18, 16, 13, 14, 10, 8, 7, 4, 2, 0],
                         list(trans[-10:]))

    def test_transitions_eca_index(self):
        ca = ECA(30)
        landscape = Landscape(ca, size=3, index=1)
        self.assertEqual(ca, landscape.network)
        self.assertEqual(3, landscape.size)
        self.assertEqual([0, 3, 2, 1, 6, 5, 6, 5], list(landscape.transitions))

        landscape = Landscape(ca, size=3, index=0)
        self.assertEqual(ca, landscape.network)
        self.assertEqual(3, landscape.size)
        self.assertEqual([0, 1, 3, 3, 5, 4, 6, 6], list(landscape.transitions))

        landscape = Landscape(ca, size=3, index=None)
        self.assertEqual(ca, landscape.network)
        self.assertEqual(3, landscape.size)
        self.assertEqual([0, 7, 7, 1, 7, 4, 2, 0], list(landscape.transitions))

    def test_transitions_eca_pin(self):
        ca = ECA(30)
        landscape = Landscape(ca, size=3, pin=[1])
        self.assertEqual(ca, landscape.network)
        self.assertEqual(3, landscape.size)
        self.assertEqual([0, 5, 7, 3, 5, 4, 2, 2], list(landscape.transitions))

        landscape = Landscape(ca, size=3, pin=[0])
        self.assertEqual(ca, landscape.network)
        self.assertEqual(3, landscape.size)
        self.assertEqual([0, 7, 6, 1, 6, 5, 2, 1], list(landscape.transitions))

        landscape = Landscape(ca, size=3, pin=None)
        self.assertEqual(ca, landscape.network)
        self.assertEqual(3, landscape.size)
        self.assertEqual([0, 7, 7, 1, 7, 4, 2, 0], list(landscape.transitions))

    def test_transitions_eca_values(self):
        ca = ECA(30)
        landscape = Landscape(ca, size=3, values={0: 1})
        self.assertEqual(ca, landscape.network)
        self.assertEqual(3, landscape.size)
        self.assertEqual([1, 7, 7, 1, 7, 5, 3, 1], list(landscape.transitions))

        landscape = Landscape(ca, size=3, values={1: 0})
        self.assertEqual(ca, landscape.network)
        self.assertEqual(3, landscape.size)
        self.assertEqual([0, 5, 5, 1, 5, 4, 0, 0], list(landscape.transitions))

        landscape = Landscape(ca, size=3, values={})
        self.assertEqual(ca, landscape.network)
        self.assertEqual(3, landscape.size)
        self.assertEqual([0, 7, 7, 1, 7, 4, 2, 0], list(landscape.transitions))

    def test_transitions_wtnetwork(self):
        net = WTNetwork(
            weights=[[1, 0], [-1, 1]],
            thresholds=[0.5, 0.0],
            theta=WTNetwork.positive_threshold
        )

        landscape = Landscape(net)
        self.assertEqual(net, landscape.network)
        self.assertEqual(2, landscape.size)
        self.assertEqual([2, 1, 2, 3], list(landscape.transitions))

    def test_transitions_wtnetwork_index(self):
        net = WTNetwork(
            weights=[[1, 0], [-1, 1]],
            thresholds=[0.5, 0.0],
            theta=WTNetwork.positive_threshold
        )

        landscape = Landscape(net, index=1)
        self.assertEqual(net, landscape.network)
        self.assertEqual(2, landscape.size)
        self.assertEqual([2, 1, 2, 3], list(landscape.transitions))

        landscape = Landscape(net, index=0)
        self.assertEqual(net, landscape.network)
        self.assertEqual(2, landscape.size)
        self.assertEqual([0, 1, 2, 3], list(landscape.transitions))

        landscape = Landscape(net, index=None)
        self.assertEqual(net, landscape.network)
        self.assertEqual(2, landscape.size)
        self.assertEqual([2, 1, 2, 3], list(landscape.transitions))

    def test_transitions_wtnetwork_pin(self):
        net = WTNetwork(
            weights=[[1, 0], [-1, 1]],
            thresholds=[0.5, 0.0],
            theta=WTNetwork.positive_threshold
        )

        landscape = Landscape(net, pin=[1])
        self.assertEqual(net, landscape.network)
        self.assertEqual(2, landscape.size)
        self.assertEqual([0, 1, 2, 3], list(landscape.transitions))

        landscape = Landscape(net, pin=[0])
        self.assertEqual(net, landscape.network)
        self.assertEqual(2, landscape.size)
        self.assertEqual([2, 1, 2, 3], list(landscape.transitions))

        landscape = Landscape(net, pin=None)
        self.assertEqual(net, landscape.network)
        self.assertEqual(2, landscape.size)
        self.assertEqual([2, 1, 2, 3], list(landscape.transitions))

    def test_transitions_wtnetwork_values(self):
        net = WTNetwork(
            weights=[[1, 0], [-1, 1]],
            thresholds=[0.5, 0.0],
            theta=WTNetwork.positive_threshold
        )

        landscape = Landscape(net, values={0: 1})
        self.assertEqual(net, landscape.network)
        self.assertEqual(2, landscape.size)
        self.assertEqual([3, 1, 3, 3], list(landscape.transitions))

        landscape = Landscape(net, values={1: 0})
        self.assertEqual(net, landscape.network)
        self.assertEqual(2, landscape.size)
        self.assertEqual([0, 1, 0, 1], list(landscape.transitions))

        landscape = Landscape(net, values={})
        self.assertEqual(net, landscape.network)
        self.assertEqual(2, landscape.size)
        self.assertEqual([2, 1, 2, 3], list(landscape.transitions))

        # Add more test for different thetas?

    def test_transitions_logicnetwork(self):
        net = LogicNetwork([((1,), {'0', '1'}), ((0,), {'1'})])

        landscape = Landscape(net)
        self.assertEqual(net, landscape.network)
        self.assertEqual(2, landscape.size)
        self.assertEqual([1, 3, 1, 3], list(landscape.transitions))

    def test_transitions_logicnetwork_index(self):
        net = LogicNetwork([((1,), {'0', '1'}), ((0,), {'1'})])

        landscape = Landscape(net, index=1)
        self.assertEqual(net, landscape.network)
        self.assertEqual(2, landscape.size)
        self.assertEqual([0, 3, 0, 3], list(landscape.transitions))

        landscape = Landscape(net, index=0)
        self.assertEqual(net, landscape.network)
        self.assertEqual(2, landscape.size)
        self.assertEqual([1, 1, 3, 3], list(landscape.transitions))

        landscape = Landscape(net, index=None)
        self.assertEqual(net, landscape.network)
        self.assertEqual(2, landscape.size)
        self.assertEqual([1, 3, 1, 3], list(landscape.transitions))

    def test_transitions_logicnetwork_pin(self):
        net = LogicNetwork([((1,), {'0', '1'}), ((0,), {'1'})])

        landscape = Landscape(net, pin=[1])
        self.assertEqual(net, landscape.network)
        self.assertEqual(2, landscape.size)
        self.assertEqual([1, 1, 3, 3], list(landscape.transitions))

        landscape = Landscape(net, pin=[0])
        self.assertEqual(net, landscape.network)
        self.assertEqual(2, landscape.size)
        self.assertEqual([0, 3, 0, 3], list(landscape.transitions))

        landscape = Landscape(net, pin=None)
        self.assertEqual(net, landscape.network)
        self.assertEqual(2, landscape.size)
        self.assertEqual([1, 3, 1, 3], list(landscape.transitions))

    def test_transitions_logicnetwork_values(self):
        net = LogicNetwork([((1,), {'0', '1'}), ((0,), {'1'})])

        landscape = Landscape(net, values={0: 1})
        self.assertEqual(net, landscape.network)
        self.assertEqual(2, landscape.size)
        self.assertEqual([1, 3, 1, 3], list(landscape.transitions))

        landscape = Landscape(net, values={1: 0})
        self.assertEqual(net, landscape.network)
        self.assertEqual(2, landscape.size)
        self.assertEqual([1, 1, 1, 1], list(landscape.transitions))

        landscape = Landscape(net, values={})
        self.assertEqual(net, landscape.network)
        self.assertEqual(2, landscape.size)
        self.assertEqual([1, 3, 1, 3], list(landscape.transitions))

        # Add more test for different thetas?

    def test_transitions_spombe(self):
        landscape = Landscape(s_pombe)
        self.assertEqual(s_pombe, landscape.network)
        self.assertEqual(9, landscape.size)
        trans = landscape.transitions
        self.assertEqual(512, len(trans))
        self.assertEqual([2, 2, 130, 130, 4, 0, 128,
                          128, 8, 0], list(trans[:10]))
        self.assertEqual([464, 464, 344, 336, 464, 464, 348,
                          336, 464, 464], list(trans[-10:]))

    def test_is_state_space(self):
        self.assertTrue(issubclass(Landscape, StateSpace))

        space = s_pombe.state_space()
        landscape = Landscape(s_pombe)
        self.assertEqual(list(space), list(landscape))
        self.assertEqual([space.encode(state) for state in space],
                         list(map(landscape.encode, landscape)))

        ca = ECA(30)
        space = ca.state_space(10)
        landscape = Landscape(ca, size=10)
        self.assertEqual(list(space), list(landscape))
        self.assertEqual([space.encode(state) for state in space],
                         list(map(landscape.encode, landscape)))

    def test_attractors_eca(self):
        networks = [(ECA(30), 2, 3), (ECA(30), 3, 1), (ECA(30), 4, 4),
                    (ECA(30), 5, 2), (ECA(30), 6, 3), (ECA(110), 2, 1),
                    (ECA(110), 3, 1), (ECA(110), 4, 3), (ECA(110), 5, 1),
                    (ECA(110), 6, 3)]

        for rule, width, size in networks:
            landscape = Landscape(rule, size=width)
            self.assertEqual(size, len(landscape.attractors))

    def test_basins_eca(self):
        networks = [(ECA(30), 2, 3), (ECA(30), 3, 1), (ECA(30), 4, 4),
                    (ECA(30), 5, 2), (ECA(30), 6, 3), (ECA(110), 2, 1),
                    (ECA(110), 3, 1), (ECA(110), 4, 3), (ECA(110), 5, 1),
                    (ECA(110), 6, 3)]

        for rule, width, size in networks:
            landscape = Landscape(rule, size=width)
            self.assertEqual(landscape.volume, len(landscape.basins))
            self.assertEqual(size, 1 + np.max(landscape.basins))

            unique = list(set(landscape.basins))
            unique.sort()
            self.assertEqual(list(range(size)), unique)

    def test_attractors_wtnetworks(self):
        networks = [(s_pombe, 13), (s_cerevisiae, 7), (c_elegans, 5)]
        for net, size in networks:
            landscape = Landscape(net)
            self.assertEqual(size, len(landscape.attractors))
            self.assertEqual(landscape.volume, len(landscape.basins))
            self.assertEqual(size, 1 + np.max(landscape.basins))

            unique = list(set(landscape.basins))
            unique.sort()
            self.assertEqual(list(range(size)), unique)

    def test_trajectory_too_short(self):
        landscape = Landscape(ECA(30), size=3)
        with self.assertRaises(ValueError):
            landscape.trajectory([0, 0, 0], timesteps=-1)

        with self.assertRaises(ValueError):
            landscape.trajectory([0, 0, 0], timesteps=0)

    def test_trajectory_eca(self):
        landscape = Landscape(ECA(30), size=3)

        with self.assertRaises(ValueError):
            landscape.trajectory([])

        with self.assertRaises(ValueError):
            landscape.trajectory([0, 1])

        xs = [0, 1, 0]

        got = landscape.trajectory(xs)
        self.assertEqual([0, 1, 0], xs)
        self.assertEqual([[0, 1, 0], [1, 1, 1], [0, 0, 0]], got)

        got = landscape.trajectory(xs, timesteps=5)
        self.assertEqual([0, 1, 0], xs)
        self.assertEqual([[0, 1, 0], [1, 1, 1], [0, 0, 0],
                          [0, 0, 0], [0, 0, 0], [0, 0, 0]], got)

        got = landscape.trajectory(xs, encode=False)
        self.assertEqual([0, 1, 0], xs)
        self.assertEqual([[0, 1, 0], [1, 1, 1], [0, 0, 0]], got)

        got = landscape.trajectory(xs, timesteps=5, encode=False)
        self.assertEqual([0, 1, 0], xs)
        self.assertEqual([[0, 1, 0], [1, 1, 1], [0, 0, 0],
                          [0, 0, 0], [0, 0, 0], [0, 0, 0]], got)

        got = landscape.trajectory(xs, encode=True)
        self.assertEqual([0, 1, 0], xs)
        self.assertEqual([2, 7, 0], list(got))

        got = landscape.trajectory(xs, timesteps=5, encode=True)
        self.assertEqual([0, 1, 0], xs)
        self.assertEqual([2, 7, 0, 0, 0, 0], list(got))

        xs = 2
        got = landscape.trajectory(xs)
        self.assertEqual([2, 7, 0], list(got))

        got = landscape.trajectory(xs, timesteps=5)
        self.assertEqual([2, 7, 0, 0, 0, 0], list(got))

        got = landscape.trajectory(xs, encode=True)
        self.assertEqual([2, 7, 0], list(got))

        got = landscape.trajectory(xs, timesteps=5, encode=True)
        self.assertEqual([2, 7, 0, 0, 0, 0], list(got))

        got = landscape.trajectory(xs, encode=False)
        self.assertEqual([[0, 1, 0], [1, 1, 1], [0, 0, 0]], got)

        got = landscape.trajectory(xs, timesteps=5, encode=False)
        self.assertEqual([[0, 1, 0], [1, 1, 1], [0, 0, 0],
                          [0, 0, 0], [0, 0, 0], [0, 0, 0]], got)

    def test_trajectory_wtnetwork(self):
        net = WTNetwork(
            weights=[[1, 0], [-1, 0]],
            thresholds=[0.5, 0.0],
            theta=WTNetwork.positive_threshold
        )

        landscape = Landscape(net)

        state = [0, 0]
        got = landscape.trajectory(state)
        self.assertEqual([0, 0], state)
        self.assertEqual([[0, 0], [0, 1]], got)

        got = landscape.trajectory(state, timesteps=3)
        self.assertEqual([0, 0], state)
        self.assertEqual([[0, 0], [0, 1], [0, 1], [0, 1]], got)

        got = landscape.trajectory(state, encode=False)
        self.assertEqual([0, 0], state)
        self.assertEqual([[0, 0], [0, 1]], got)

        got = landscape.trajectory(state, timesteps=3, encode=False)
        self.assertEqual([0, 0], state)
        self.assertEqual([[0, 0], [0, 1], [0, 1], [0, 1]], got)

        got = landscape.trajectory(state, encode=True)
        self.assertEqual([0, 0], state)
        self.assertEqual([0, 2], got)

        got = landscape.trajectory(state, timesteps=3, encode=True)
        self.assertEqual([0, 0], state)
        self.assertEqual([0, 2, 2, 2], got)

        state = 0
        got = landscape.trajectory(state)
        self.assertEqual([0, 2], got)

        got = landscape.trajectory(state, timesteps=3)
        self.assertEqual([0, 2, 2, 2], got)

        got = landscape.trajectory(state, encode=True)
        self.assertEqual([0, 2], got)

        got = landscape.trajectory(state, timesteps=3, encode=True)
        self.assertEqual([0, 2, 2, 2], got)

        got = landscape.trajectory(state, encode=False)
        self.assertEqual([[0, 0], [0, 1]], got)

        got = landscape.trajectory(state, timesteps=3, encode=False)
        self.assertEqual([[0, 0], [0, 1], [0, 1], [0, 1]], got)

    def test_trajectory_logicnetwork(self):
        net = LogicNetwork([((1, 2), {'01', '10'}),
                            ((0, 2), {'01', '10', '11'}),
                            ((0, 1), {'11'})])

        landscape = Landscape(net)

        state = [0, 1, 0]
        got = landscape.trajectory(state, encode=False)
        self.assertEqual([0, 1, 0], state)
        self.assertEqual([[0, 1, 0], [1, 0, 0]], got)

        got = landscape.trajectory(state, timesteps=3, encode=False)
        self.assertEqual([0, 1, 0], state)
        self.assertEqual([[0, 1, 0], [1, 0, 0], [0, 1, 0], [1, 0, 0]], got)

        got = landscape.trajectory(state)
        self.assertEqual([0, 1, 0], state)
        self.assertEqual([[0, 1, 0], [1, 0, 0]], got)

        got = landscape.trajectory(state, timesteps=3)
        self.assertEqual([0, 1, 0], state)
        self.assertEqual([[0, 1, 0], [1, 0, 0], [0, 1, 0], [1, 0, 0]], got)

        got = landscape.trajectory(state, encode=True)
        self.assertEqual([0, 1, 0], state)
        self.assertEqual([2, 1], got)

        got = landscape.trajectory(state, timesteps=3, encode=True)
        self.assertEqual([0, 1, 0], state)
        self.assertEqual([2, 1, 2, 1], got)

        state = 2
        got = landscape.trajectory(state)
        self.assertEqual([2, 1], got)

        got = landscape.trajectory(state, timesteps=3)
        self.assertEqual([2, 1, 2, 1], got)

        got = landscape.trajectory(state, encode=True)
        self.assertEqual([2, 1], got)

        got = landscape.trajectory(state, timesteps=3, encode=True)
        self.assertEqual([2, 1, 2, 1], got)

        got = landscape.trajectory(state, encode=False)
        self.assertEqual([[0, 1, 0], [1, 0, 0]], got)

        got = landscape.trajectory(state, timesteps=3, encode=False)
        self.assertEqual([[0, 1, 0], [1, 0, 0], [0, 1, 0], [1, 0, 0]], got)

    def test_timeseries_too_short(self):
        landscape = Landscape(ECA(30), size=3)
        with self.assertRaises(ValueError):
            landscape.timeseries(-1)

        with self.assertRaises(ValueError):
            landscape.timeseries(0)

    def test_timeseries_eca(self):
        rule = ECA(30)
        for size in [5, 7, 11]:
            landscape = Landscape(rule, size=size)
            time = 10
            series = landscape.timeseries(time)
            self.assertEqual((size, 2**size, time + 1), series.shape)
            for index, state in enumerate(rule.state_space(size)):
                traj = landscape.trajectory(state, timesteps=time)
                for t, expect in enumerate(traj):
                    got = series[:, index, t]
                    self.assertTrue(np.array_equal(expect, got))

    def test_timeseries_wtnetworks(self):
        for net, size in [(s_pombe, 9), (s_cerevisiae, 11), (c_elegans, 8)]:
            landscape = Landscape(net)
            time = 10
            series = landscape.timeseries(time)
            self.assertEqual((size, 2**size, time + 1), series.shape)
            for index, state in enumerate(net.state_space()):
                traj = landscape.trajectory(state, timesteps=time)
                for t, expect in enumerate(traj):
                    got = series[:, index, t]
                    self.assertTrue(np.array_equal(expect, got))

    def test_basin_sizes(self):
        for net in [s_pombe, s_cerevisiae, c_elegans]:
            landscape = Landscape(net)
            basins = landscape.basins
            histogram = [0] * (np.max(basins) + 1)
            for b in basins:
                histogram[b] += 1
            self.assertEqual(histogram, list(landscape.basin_sizes))

    def test_basin_entropy_eca(self):
        networks = [(ECA(30),  2, 1.500000, 0.451545),
                    (ECA(30),  3, 0.000000, 0.000000),
                    (ECA(30),  4, 1.186278, 0.357105),
                    (ECA(30),  5, 0.337290, 0.101534),
                    (ECA(30),  6, 0.231872, 0.069801),
                    (ECA(110), 2, 0.000000, 0.000000),
                    (ECA(110), 3, 0.000000, 0.000000),
                    (ECA(110), 4, 1.561278, 0.469992),
                    (ECA(110), 5, 0.000000, 0.000000),
                    (ECA(110), 6, 1.469012, 0.442217)]

        for net, width, base2, base10 in networks:
            landscape = Landscape(net, size=width)
            self.assertAlmostEqual(base2, landscape.basin_entropy(), places=6)
            self.assertAlmostEqual(
                base2, landscape.basin_entropy(base=2), places=6)
            self.assertAlmostEqual(
                base10, landscape.basin_entropy(base=10), places=6)

    def test_basin_entropy_wtnetwork(self):
        networks = [(s_pombe,      1.221889, 0.367825),
                    (s_cerevisiae, 0.783858, 0.235965),
                    (c_elegans,    0.854267, 0.257160)]

        for net, base2, base10 in networks:
            landscape = Landscape(net)
            self.assertAlmostEqual(base2, landscape.basin_entropy(), places=6)
            self.assertAlmostEqual(
                base2, landscape.basin_entropy(base=2), places=6)
            self.assertAlmostEqual(
                base10, landscape.basin_entropy(base=10), places=6)

    def test_graph_eca(self):
        for size in range(2, 7):
            landscape = Landscape(ECA(30), size=size)
            g = landscape.graph
            self.assertEqual(landscape.volume, g.number_of_nodes())
            self.assertEqual(landscape.volume, g.number_of_edges())

    def test_graph_wtnetworks(self):
        for net in [s_pombe, s_cerevisiae, c_elegans]:
            landscape = Landscape(net)
            g = landscape.graph
            self.assertEqual(landscape.volume, g.number_of_nodes())
            self.assertEqual(landscape.volume, g.number_of_edges())

    def test_in_degree(self):
        for code in [30, 110, 21, 43]:
            for size in range(2, 7):
                landscape = Landscape(ECA(code), size=size)
                in_degrees = np.empty(landscape.volume, dtype=np.int)
                for i in range(landscape.volume):
                    in_degrees[i] = np.count_nonzero(
                        landscape.transitions == i)
                self.assertEqual(list(in_degrees), list(landscape.in_degrees))

        for net in [s_pombe, s_cerevisiae, c_elegans]:
            landscape = Landscape(net)
            in_degrees = np.empty(landscape.volume, dtype=np.int)
            for i in range(landscape.volume):
                in_degrees[i] = np.count_nonzero(landscape.transitions == i)
            self.assertEqual(list(in_degrees), list(landscape.in_degrees))

    def test_heights(self):
        for code in [30, 110, 21, 43]:
            for size in range(2, 7):
                landscape = Landscape(ECA(code), size=size)
                heights = [0] * landscape.volume
                for i in range(landscape.volume):
                    b = landscape.basins[i]
                    state = i
                    while state not in landscape.attractors[b]:
                        heights[i] += 1
                        state = landscape.transitions[state]
                self.assertEqual(heights, list(landscape.heights))

        for net in [s_pombe, s_cerevisiae, c_elegans]:
            landscape = Landscape(net)
            heights = [0] * landscape.volume
            for i in range(landscape.volume):
                b = landscape.basins[i]
                state = i
                while state not in landscape.attractors[b]:
                    heights[i] += 1
                    state = landscape.transitions[state]
            self.assertEqual(heights, list(landscape.heights))

    def test_attractor_lengths(self):
        for code in [30, 110, 21, 43]:
            for size in range(2, 7):
                landscape = Landscape(ECA(code), size=size)
                lengths = list(map(len, landscape.attractors))
                self.assertEqual(lengths, list(landscape.attractor_lengths))

        for net in [s_pombe, s_cerevisiae, c_elegans]:
            landscape = Landscape(net)
            lengths = list(map(len, landscape.attractors))
            self.assertEqual(lengths, list(landscape.attractor_lengths))

    def test_recurrence_times(self):
        for code in [30, 110, 21, 43]:
            for size in range(2, 7):
                landscape = Landscape(ECA(code), size=size)
                recurrence_times = [0] * landscape.volume
                for i in range(landscape.volume):
                    b = landscape.basins[i]
                    recurrence_times[i] = landscape.heights[i] + \
                        landscape.attractor_lengths[b] - 1
                self.assertEqual(recurrence_times, list(
                    landscape.recurrence_times))

        for net in [s_pombe, s_cerevisiae, c_elegans]:
            landscape = Landscape(net)
            recurrence_times = [0] * landscape.volume
            for i in range(landscape.volume):
                b = landscape.basins[i]
                recurrence_times[i] = landscape.heights[i] + \
                    landscape.attractor_lengths[b] - 1
            self.assertEqual(recurrence_times, list(
                landscape.recurrence_times))
